r"""Components for computing derivatives of the Eikonax solver.

This module contains two main components. Firstly, the
[`PartialDerivator`][eikonax.derivator.PartialDerivator] evaluates the partial derivatives of the
global Eikonax update operator $\mathbf{G}$ w.r.t. the parameter tensor field $\mathbf{M}$ and the
corresponding solution vector $\mathbf{u}$ obtained from a forward solve. The
[`DerivativeSolver`][eikonax.derivator.DerivativeSolver] component makes use of the fixed
point/adjoint property of the Eikonax solver to evaluate total parametric derivatives.

Classes:
    PartialDerivatorData: Settings for initialization of partial derivator
    PartialDerivator: Component for computing partial derivatives of the Godunov Update operator
    DerivativeSolver: Main component for obtaining gradients from partial derivatives
"""

from dataclasses import dataclass
from numbers import Real
from typing import Annotated

import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
import numpy.typing as npt
import scipy as sp
import sparse as spa
from beartype.vale import Is
from jaxtyping import Float as jtFloat
from jaxtyping import Int as jtInt

from . import corefunctions, preprocessing


@dataclass
class PartialDerivatorData:
    """Settings for initialization of partial derivator.

    See the [Forward Solver](../usage/solve.md) documentation for more detailed explanations.

    Attributes:
        softminmax_order (int): Order of the the soft minmax function for differentiable
            transformation of the update parameters
        softminmax_cutoff (Real): Cut-off in for minmax transformation, beyond which zero
            sensitivity is assumed.
    """

    use_soft_update: bool
    softminmax_order: Annotated[int, Is[lambda x: x > 0]]
    softminmax_cutoff: Annotated[Real, Is[lambda x: x > 0]]


# ==================================================================================================
class PartialDerivator(eqx.Module):
    r"""Component for computing partial derivatives of the Godunov Update operator.

    Given a tensor field $M$ and a solution vector $u$, the partial derivator computes the partial
    derivatives of the global Eikonax update operator with respect to the solution vector,
    $\mathbf{G}_u(\mathbf{u}, \mathbf{m})$, and the tensor field,
    $\mathbf{G}_M(\mathbf{u}, \mathbf{M})$. All derivatives are computed on the vertex level,
    exploiting the locality of interactions in the update operator (only adjacent simplices are
    considered). Therefore, we can indeed assemble the complete derivative operators as parse data
    structures, not just Jacobian-vector or vector-Jacobian products, within a single pass over the
    computational mesh. Atomic functions on the vertex level are differentiated with Jax.

    !!! info
        For the computation of the derivatives, so-called 'self-updates' are disabled. These updates
        occur when a vertex does not receive a lower update value from any direction than the value
        it currently has. At the correct solution point, this case cannot occur due to the causality
        of the update stencil.

    Methods:
        compute_partial_derivatives: Compute the partial derivatives of the Godunov update operator
            with respect to the solution vector and the parameter tensor field, given a state for
            both variables
    """

    # Equinox modules are data classes, so we need to specify attributes on class level
    _num_vertices: int
    _num_simplices: int
    _vertices: jtFloat[jax.Array, "num_vertices dim"]
    _adjacency_data: jtInt[jax.Array, "num_vertices max_num_adjacent_simplices 4"]
    _initial_site_inds: jtInt[jax.Array, "num_initial_sites"]
    _initial_site_values: jtFloat[jax.Array | npt.NDArray, "num_initial_sites"]
    _use_soft_update: bool
    _softminmax_order: int
    _softminmax_cutoff: int

    # ----------------------------------------------------------------------------------------------
    def __init__(
        self,
        mesh_data: preprocessing.MeshData,
        derivator_data: PartialDerivatorData,
        initial_sites: preprocessing.InitialSites,
    ) -> None:
        """Constructor for the partial derivator object.

        Args:
            mesh_data (preprocessing.MeshData): Mesh data object also utilized for the Eikonax
                solver, contains adjacency data for every vertex.
            derivator_data (PartialDerivatorData): Settings for initialization of the derivator.
            initial_sites (preprocessing.InitialSites): Locations and values at source points
        """
        self._num_vertices = mesh_data.num_vertices
        self._num_simplices = mesh_data.num_simplices
        self._vertices = mesh_data.vertices
        self._adjacency_data = mesh_data.adjacency_data
        self._initial_site_inds = initial_sites.inds
        self._initial_site_values = initial_sites.values
        self._use_soft_update = derivator_data.use_soft_update
        self._softminmax_order = derivator_data.softminmax_order
        self._softminmax_cutoff = derivator_data.softminmax_cutoff

    # ----------------------------------------------------------------------------------------------
    def compute_partial_derivatives(
        self,
        solution_vector: jtFloat[jax.Array | npt.NDArray, "num_vertices"],
        tensor_field: jtFloat[jax.Array | npt.NDArray, "num_simplices dim dim"],
    ) -> tuple[spa.COO, spa.COO]:
        r"""Compute the partial derivatives of the Godunov update operator.

        This method provides the main interface for computing the partial derivatives of the global
        Eikonax update operator with respect to the solution vector and the parameter tensor field.
        The updates are computed locally for each vertex, such that the resulting data structures
        are sparse. Subsequently, further zero entries are removed to reduce the memory footprint.
        The derivatives computed in this component can be utilized to compute the total parametric
        derivative via a fix point equation, given that the provided solution vector is that
        fix point.
        The computation of partial derivatives is possible with a single pass over the mesh, since
        the solution of the Eikonax equation, and therefore causality within the Godunov update
        scheme, is known.

        !!! note
            The derivator expects the metric tensor field as used in the inner product for the
            update stencil of the eikonal equation. This is the **INVERSE** of the conductivity
            tensor, which is the actual tensor field in the eikonal equation. The
            [`Tensorfield`][eikonax.tensorfield.TensorField] component provides the inverse tensor
            field.

        Args:
            solution_vector (jax.Array): Current solution
            tensor_field (jax.Array): Parameter field

        Returns:
            tuple[spa.COO, spa.COO]:
                Partial derivatives $\mathbf{G}_u$ and $\mathbf{G}_M$, both returned as
                [sparse COO](https://sparse.pydata.org/en/stable/api/COO/) matrices. $\mathbf{G}_u$
                has shape $N_V \times N_V$, and $\mathbf{G}_M$ has shape
                $N_V \times N_S \times d \times d$.
        """
        solution_vector = jnp.array(solution_vector, dtype=jnp.float32)
        tensor_field = jnp.array(tensor_field, dtype=jnp.float32)
        partial_derivative_solution, partial_derivative_parameter = (
            self._compute_global_partial_derivatives(
                solution_vector,
                tensor_field,
            )
        )
        sparse_partial_solution = self._compress_partial_derivative_solution(
            partial_derivative_solution
        )
        sparse_partial_parameter = self._compress_partial_derivative_parameter(
            partial_derivative_parameter
        )
        return sparse_partial_solution, sparse_partial_parameter

    # ----------------------------------------------------------------------------------------------
    def _compress_partial_derivative_solution(
        self,
        partial_derivative_solution: jtFloat[
            jax.Array, "num_vertices max_num_adjacent_simplices 2"
        ],
    ) -> spa.COO:
        r"""Compress the partial derivative data with respect to the solution vector.

        Compression consists of three steps:

        1. Remove zero entries in the sensitivity vector
        2. Set the sensitivity vector to zero at the initial sites, but keep them for later
           computations.
        3. Convert arrays to sparse COO formmat.

        Args:
            partial_derivative_solution (jax.Array): Raw data from partial derivative computation,
                with shape `(N, num_adjacent_simplices, 2)`.

        Returns:
            spa.COO: Matrix $\mathbf{G}_u$ as above
        """
        current_inds = self._adjacency_data[:, 0, 0]
        adjacent_inds = self._adjacency_data[:, :, 1:3]

        nonzero_mask = jnp.nonzero(partial_derivative_solution)
        rows_compressed = current_inds[nonzero_mask[0]]
        columns_compressed = adjacent_inds[nonzero_mask]
        values_compressed = partial_derivative_solution[nonzero_mask]

        initial_site_mask = jnp.where(rows_compressed == self._initial_site_inds)
        values_compressed = values_compressed.at[initial_site_mask].set(
            jnp.zeros(self._initial_site_inds.shape)
        )
        sparse_coo_matrix = spa.COO(
            coords=(rows_compressed, columns_compressed),
            data=values_compressed,
            shape=(self._num_vertices, self._num_vertices),
        )
        return sparse_coo_matrix

    # ----------------------------------------------------------------------------------------------
    def _compress_partial_derivative_parameter(
        self,
        partial_derivative_parameter: jtFloat[
            jax.Array, "num_vertices max_num_adjacent_simplices dim dim"
        ],
    ) -> spa.COO:
        r"""Compress the partial derivative data with respect to the parameter tensor field.

        Compression consists of three steps:

        1. Remove tensor components from the sensitivity data, if all entries are zero
        2. Set the sensitivity vector to zero at the initial sites, but keep them for later
           computations.
        3. Convert arrays to sparse COO format.

        Args:
            partial_derivative_parameter (jax.Array): Raw data from partial derivative
                computation, with shape `(N, num_adjacent_simplices, dim, dim)`.

        Returns:
            spa.COO: Matrix $\mathbf{G}_M$ as above
        """
        vertex_inds = self._adjacency_data[:, 0, 0]
        simplex_inds = self._adjacency_data[:, :, 3]
        tensor_dim = partial_derivative_parameter.shape[2]
        tensor_inds = jnp.arange(tensor_dim, dtype=jnp.int32)

        nonzero_mask = jnp.nonzero(partial_derivative_parameter)
        rows_compressed = vertex_inds[nonzero_mask[0]]
        simplices_compressed = simplex_inds[(nonzero_mask[0], nonzero_mask[1])]
        tensor_d1_compressed = tensor_inds[nonzero_mask[2]]
        tensor_d2_compressed = tensor_inds[nonzero_mask[3]]
        values_compressed = partial_derivative_parameter[nonzero_mask]

        initial_site_mask = jnp.where(rows_compressed == self._initial_site_inds)
        values_compressed = values_compressed.at[initial_site_mask].set(
            jnp.zeros(initial_site_mask[0].size)
        )
        sparse_coo_matrix = spa.COO(
            coords=(
                rows_compressed,
                simplices_compressed,
                tensor_d1_compressed,
                tensor_d2_compressed,
            ),
            data=values_compressed,
            shape=(self._num_vertices, self._num_simplices, tensor_dim, tensor_dim),
        )

        return sparse_coo_matrix

    # ----------------------------------------------------------------------------------------------
    @eqx.filter_jit
    def _compute_global_partial_derivatives(
        self,
        solution_vector: jtFloat[jax.Array, "num_vertices"],
        tensor_field: jtFloat[jax.Array, "num_simplices dim dim"],
    ) -> tuple[
        jtFloat[jax.Array, "num_vertices max_num_adjacent_simplices 2"],
        jtFloat[jax.Array, "num_vertices max_num_adjacent_simplices dim dim"],
    ]:
        """Compute partial derivatives of the global update operator.

        The method is a jitted and vectorized call to the
        [`_compute_vertex_partial_derivative`][eikonax.derivator.PartialDerivator._compute_vertex_partial_derivatives]
        method.

        Args:
            solution_vector (jax.Array): Global solution vector
            tensor_field (jax.Array): Global parameter tensor field

        Returns:
            tuple[jax.Array, jax.Array]: Raw data for partial derivatives, with shapes
                (N, num_adjacent_simplices, 2) and (N, num_adjacent_simplices, dim, dim), N depends
                on the number of identical update paths for the vertices in the mesh.
        """
        global_partial_derivative_function = jax.vmap(
            self._compute_vertex_partial_derivatives,
            in_axes=(None, None, 0),
        )
        partial_derivative_solution, partial_derivative_parameter = (
            global_partial_derivative_function(solution_vector, tensor_field, self._adjacency_data)
        )
        return partial_derivative_solution, partial_derivative_parameter

    # ----------------------------------------------------------------------------------------------
    def _compute_vertex_partial_derivatives(
        self,
        solution_vector: jtFloat[jax.Array, "num_vertices"],
        tensor_field: jtFloat[jax.Array, "num_simplices dim dim"],
        adjacency_data: jtInt[jax.Array, "max_num_adjacent_simplices 4"],
    ) -> tuple[
        jtFloat[jax.Array, "max_num_adjacent_simplices 2"],
        jtFloat[jax.Array, "max_num_adjacent_simplices dim dim"],
    ]:
        """Compute partial derivatives for the update of a single vertex.

        The method computes candidates for all respective subterms through calls to further methods.
        These candidates are filtered for feasibility by means of JAX filters.
        The sofmin function (and its gradient) is applied to the directions of all optimal
        updates to ensure differentiability, other contributions are discarded.
        Lasty, the evaluated contributions are combined according to the form of the
        "total differential" for the partial derivatives.

        Args:
            solution_vector (jax.Array): Global solution vector
            tensor_field (jax.Array): Global parameter tensor field
            adjacency_data (jax.Array): Adjacency data for the vertex under consideration

        Returns:
            tuple[jax.Array, jax.Array]: Partial derivatives for the given vertex
        """
        max_num_adjacent_simplices = adjacency_data.shape[0]
        tensor_dim = tensor_field.shape[1]
        assert adjacency_data.shape == (max_num_adjacent_simplices, 4), (
            f"node-level adjacency data needs to have shape ({max_num_adjacent_simplices}, 4), "
            f"but has shape {adjacency_data.shape}"
        )

        vertex_update_candidates = corefunctions.compute_vertex_update_candidates(
            solution_vector,
            tensor_field,
            adjacency_data,
            self._vertices,
            self._use_soft_update,
            self._softminmax_order,
            self._softminmax_cutoff,
        )
        grad_update_solution_candidates, grad_update_parameter_candidates = (
            self._compute_vertex_partial_derivative_candidates(
                solution_vector, tensor_field, adjacency_data
            )
        )
        min_value, grad_update_solution_candidates, grad_update_parameter_candidates = (
            self._filter_candidates(
                vertex_update_candidates,
                grad_update_solution_candidates,
                grad_update_parameter_candidates,
            )
        )
        average_grad = corefunctions.grad_average(
            vertex_update_candidates.flatten(), min_value
        ).reshape(vertex_update_candidates.shape)

        grad_update_solution = jnp.zeros((max_num_adjacent_simplices, 2))
        grad_update_parameter = jnp.zeros((max_num_adjacent_simplices, tensor_dim, tensor_dim))
        for i in range(max_num_adjacent_simplices):
            grad_update_solution = grad_update_solution.at[i, :].set(
                jnp.tensordot(average_grad[i, :], grad_update_solution_candidates[i, ...], axes=1)
            )
            grad_update_parameter = grad_update_parameter.at[i, ...].set(
                jnp.tensordot(average_grad[i, :], grad_update_parameter_candidates[i, ...], axes=1)
            )

        return grad_update_solution, grad_update_parameter

    # ----------------------------------------------------------------------------------------------
    def _compute_vertex_partial_derivative_candidates(
        self,
        solution_vector: jtFloat[jax.Array, "num_vertices"],
        tensor_field: jtFloat[jax.Array, "num_simplices dim dim"],
        adjacency_data: jtInt[jax.Array, "max_num_adjacent_simplices 4"],
    ) -> tuple[
        jtFloat[jax.Array, "max_num_adjacent_simplices 4 2"],
        jtFloat[jax.Array, "max_num_adjacent_simplices 4 dim dim"],
    ]:
        """Compute partial derivatives corresponding to potential update candidates for a vertex.

        Update candidates and corresponding derivatives are computed for all adjacent simplices,
        and for all possible update parameters per simplex.

        Args:
            solution_vector (jax.Array): Global solution vector
            tensor_field (jax.Array): Global parameter field
            adjacency_data (jax.Array): Adjacency data for the given vertex

        Returns:
            tuple[jax.Array, jax.Array]: Candidates for partial derivatives
        """
        max_num_adjacent_simplices = adjacency_data.shape[0]
        tensor_dim = tensor_field.shape[1]
        grad_update_solution_candidates = jnp.zeros(
            (max_num_adjacent_simplices, 4, 2), dtype=jnp.float32
        )
        grad_update_parameter_candidates = jnp.zeros(
            (max_num_adjacent_simplices, 4, tensor_dim, tensor_dim), dtype=jnp.float32
        )

        for i, indices in enumerate(adjacency_data):
            partial_solution, partial_parameter = (
                self._compute_partial_derivative_candidates_from_adjacent_simplex(
                    solution_vector, tensor_field, indices
                )
            )
            grad_update_solution_candidates = grad_update_solution_candidates.at[i, ...].set(
                partial_solution
            )
            grad_update_parameter_candidates = grad_update_parameter_candidates.at[i, ...].set(
                partial_parameter
            )

        return grad_update_solution_candidates, grad_update_parameter_candidates

    # ----------------------------------------------------------------------------------------------
    def _compute_partial_derivative_candidates_from_adjacent_simplex(
        self,
        solution_vector: jtFloat[jax.Array, "num_vertices"],
        tensor_field: jtFloat[jax.Array, "num_simplices dim dim"],
        adjacency_data: jtInt[jax.Array, "4"],
    ) -> tuple[jtFloat[jax.Array, "4 2"], jtFloat[jax.Array, "4 dim dim"]]:
        r"""Compute partial derivatives for all update candidates within an adjacent simplex.

        The update candidates are evaluated according to the different candidates for the
        optimization parameters $\lambda$. Contributions are combined to the form of the involved
        total differentials.

        Args:
            solution_vector (jax.Array): Global solution vector
            tensor_field (jax.Array): Flobal parameter field
            adjacency_data (jax.Array): Adjacency data for the given vertex and simplex

        Returns:
            tuple[jax.Array, jax.Array]: Derivative candidate from the given simplex
        """
        i, j, k, s = adjacency_data
        tensor_dim = tensor_field.shape[1]
        solution_values = jnp.array((solution_vector[j], solution_vector[k]))
        edges = corefunctions.compute_edges(i, j, k, self._vertices)
        parameter_tensor = tensor_field[s]
        if self._use_soft_update:
            lambda_array = corefunctions.compute_optimal_update_parameters_soft(
                solution_values,
                parameter_tensor,
                edges,
                self._softminmax_order,
                self._softminmax_cutoff,
            )
        else:
            lambda_array = corefunctions.compute_optimal_update_parameters_hard(
                solution_values, parameter_tensor, edges
            )
        lambda_partial_solution, lambda_partial_parameter = self._compute_lambda_grad(
            solution_values, parameter_tensor, edges
        )
        lambda_partial_solution = jnp.concatenate((jnp.zeros((2, 2)), lambda_partial_solution))
        lambda_partial_parameter = jnp.concatenate(
            (jnp.zeros((2, tensor_dim, tensor_dim)), lambda_partial_parameter)
        )
        grad_update_solution = jnp.zeros((4, 2))
        grad_update_parameter = jnp.zeros((4, tensor_dim, tensor_dim))

        for i in range(4):
            update_partial_lambda = corefunctions.grad_update_lambda(
                solution_values, parameter_tensor, lambda_array[i], edges
            )
            update_partial_solution = corefunctions.grad_update_solution(
                solution_values, parameter_tensor, lambda_array[i], edges
            )
            update_partial_parameter = corefunctions.grad_update_parameter(
                solution_values, parameter_tensor, lambda_array[i], edges
            )
            grad_update_solution = grad_update_solution.at[i, :].set(
                update_partial_lambda * lambda_partial_solution[i, :] + update_partial_solution
            )
            grad_update_parameter = grad_update_parameter.at[i, ...].set(
                update_partial_lambda * lambda_partial_parameter[i, ...] + update_partial_parameter
            )
        return grad_update_solution, grad_update_parameter

    # ----------------------------------------------------------------------------------------------
    @staticmethod
    def _filter_candidates(
        vertex_update_candidates: jtFloat[jax.Array, "max_num_adjacent_simplices 4"],
        grad_update_solution_candidates: jtFloat[jax.Array, "max_num_adjacent_simplices 4 2"],
        grad_update_parameter_candidates: jtFloat[
            jax.Array, "max_num_adjacent_simplices 4 dim dim"
        ],
    ) -> tuple[
        jtFloat[jax.Array, ""],
        jtFloat[jax.Array, "max_num_adjacent_simplices 4 2"],
        jtFloat[jax.Array, "max_num_adjacent_simplices 4 dim dim"],
    ]:
        """Mask irrelevant derivative candidates so that they are discarded later.

        Values are masked by setting them to zero or infinity, depending on the routine in which
        they are utilized later. Partial derivatives are only relevant if the corresponding update
        corresponds to an optimal path.

        Args:
            vertex_update_candidates (jax.Array): Update candidates for a given vertex
            grad_update_solution_candidates (jax.Array): Partial derivative candidates w.r.t. the
                solution vector
            grad_update_parameter_candidates (jax.Array): Partial derivative candidates w.r.t. the
                parameter field

        Returns:
            tuple[jax.Array, jax.Array, jax.Array]: Optimal update value, masked partial
                derivatives
        """
        min_value = jnp.min(vertex_update_candidates)
        min_candidate_mask = jnp.where(vertex_update_candidates == min_value, 1, 0)
        vertex_update_candidates = jnp.where(
            min_candidate_mask == 1, vertex_update_candidates, jnp.inf
        )
        grad_update_solution_candidates = jnp.where(
            min_candidate_mask[..., None] == 1, grad_update_solution_candidates, 0
        )
        grad_update_parameter_candidates = jnp.where(
            min_candidate_mask[..., None, None] == 1, grad_update_parameter_candidates, 0
        )

        return min_value, grad_update_solution_candidates, grad_update_parameter_candidates

    # ----------------------------------------------------------------------------------------------
    def _compute_lambda_grad(
        self,
        solution_values: jtFloat[jax.Array, "2"],
        parameter_tensor: jtFloat[jax.Array, "dim dim"],
        edges: tuple[
            jtFloat[jax.Array, "dim"], jtFloat[jax.Array, "dim"], jtFloat[jax.Array, "dim"]
        ],
    ) -> tuple[jtFloat[jax.Array, "4 2"], jtFloat[jax.Array, "4 dim dim"]]:
        """Compute the partial derivatives of update parameters for a single vertex.

        This method evaluates the partial derivatives of the update parameters with respect to the
        current solution vector and the given parameter field, for a single triangle.

        Args:
            solution_values (jax.Array): Current solution values at the opposite vertices of the
                considered triangle
            parameter_tensor (jax.Array): Parameter tensor for the given triangle
            edges (tuple[jax.Array, jax.Array, jax.Array]): Edges of the considered triangle

        Returns:
            tuple[jax.Array, jax.Array]: Jacobians of the update parameters w.r.t.
                the solution vector and the parameter tensor
        """
        if self._use_soft_update:
            lambda_partial_solution = corefunctions.jac_lambda_soft_solution(
                solution_values,
                parameter_tensor,
                edges,
                self._softminmax_order,
                self._softminmax_cutoff,
            )
            lambda_partial_parameter = corefunctions.jac_lambda_soft_parameter(
                solution_values,
                parameter_tensor,
                edges,
                self._softminmax_order,
                self._softminmax_cutoff,
            )
        else:
            lambda_partial_solution = corefunctions.jac_lambda_hard_solution(
                solution_values, parameter_tensor, edges
            )
            lambda_partial_parameter = corefunctions.jac_lambda_hard_parameter(
                solution_values, parameter_tensor, edges
            )

        return lambda_partial_solution, lambda_partial_parameter


# ==================================================================================================
class DerivativeSolver:
    r"""Main component for obtaining gradients from partial derivatives.

    The Eikonax [`PartialDerivator`][eikonax.derivator.PartialDerivator] computes partial
    derivatives of the global update operator with respect
    to the solution vector, $\mathbf{G}_u$, and the parameter tensor field, $\mathbf{G}_M$.
    We exploit the fact that the obtained solution candidate from a forward solve
    $\mathbf{u}\in\mathbb{R}^{N_V}$ is, up to a given accuracy, a fixed point of the
    global update operator. We further consider the scenario of $\mathbf{M}(\mathbf{m})$ being
    dependent on some parameter $\mathbf{m}\in\mathbb{R}^M$. This means we can write $\mathbf{u}$ as
    a function of $\mathbf{m}$, obeying the relation

    $$
        \mathbf{u}(\mathbf{m}) = \mathbf{G}(\mathbf{u}(\mathbf{m}), \mathbf{M}(\mathbf{m}))
    $$

    To obtain the Jacobian
    $\mathbf{J} = \frac{d\mathbf{u}}{d\mathbf{m}}\in\mathbb{R}^{N_V\times M}$,
    we simply differentiate the fixed point relation,

    $$
        \mathbf{J} = \mathbf{G}_u\mathbf{J}
        + \overbrace{\mathbf{G}_M\frac{d\mathbf{M}(\mathbf{m})}{d\mathbf{m}}}^{\mathbf{G}_m}
        \quad\Leftrightarrow\quad \mathbf{J} = (\mathbf{I} - \mathbf{G}_u)^{-1}\mathbf{G}_m
    $$

    $\mathbf{G}_u$ and $\mathbf{G}_M$ are provided by the
    [`PartialDerivator`][eikonax.derivator.PartialDerivator], whereas
    $\frac{d\mathbf{M}}{d\mathbf{m}}$ is computed as the Jacobian of the
    [`TensorField`][eikonax.tensorfield.TensorField] component.

    We are typically not interested in the full Jacobian, but rather in the gradient of some
    cost functional $l:\mathbb{R}^{N_V}\to\mathbb{R},\ l=l(\mathbf{u}(\mathbf{m}))$ with respect to
    $\mathbf{m}$. The gradient is given as

    $$
        \mathbf{g}(\mathbf{m}) = \frac{d l}{d\mathbf{m}}
        = \overbrace{\frac{d l}{d\mathbf{u}}}^{l_u^T}\mathbf{J}
        = \overbrace{l_u^T(\mathbf{I} - \mathbf{G}_u)^{-1}}^{\mathbf{v}^T}\mathbf{G}_m.
    $$

    We can identify $\mathbf{v}$ as the adjoint variable, which is obtained by solving the linear
    **discrete adjoint equation**,

    $$
        (\mathbf{I} - \mathbf{G}_u)^T\mathbf{v} = l_u.
    $$

    *Now comes the catch*: Through the strict causality of the Godunov update operator, we can find
    a unique and consistent ordering of vertex indices, such that the solution at a vertex $i$ is
    only informed by the solution at a vertex $j$, if $j$ occurs before $i$ in that ordering. The
    matrix $\mathbf{G}_u$ has to form a directed, acyclic graph.
    This means that there is an orthogonal permutation matrix $\mathbf{P}$ such that for
    $\bar{\mathbf{G}}_u = \mathbf{P}\mathbf{G}_u\mathbf{P}^T$ an entry $(\bar{\mathbf{G}}_u)_{ij}$
    is only non-zero if $i > j$. In total, we can write

    $$
    \begin{align*}
    (\mathbf{I}-\mathbf{G}_u)^T\mathbf{v} = \mathbf{l}_u &\Leftrightarrow
    \mathbf{P}(\mathbf{I}-\mathbf{G}_u)^T\mathbf{P}^T
    \overbrace{\mathbf{P}\mathbf{v}}^{\bar{\mathbf{v}}}
    = \overbrace{\mathbf{P}\mathbf{l}_u}^{\bar{\mathbf{l}}_u} \nonumber \\
    & \Leftrightarrow \overbrace{(\mathbf{I} - \bar{\mathbf{G}}_u)^T}^{\bar{\mathbf{A}}}\bar{v}
    =\bar{\mathbf{l}}_u
    \end{align*}
    $$

    where $\bar{\mathbf{A}}$ is an upper triangular matrix with unit diagonal. Hence, it is
    invertible through simple back-substitution.

    The `DerivativeSolver` component does exactly this: It sets up the matrices $\mathbf{P}$ and
    $\bar{\mathbf{A}}$, permutates inputs/outputs, and solves the sparse linear system through
    back-substitution.

    !!! tip "Speedy gradients"
        Given a solution vector $\mathbf{u}$, Eikonax computes derivatives with linear complexity.
        Even more, for a given evaluation point, we can evaluate an arbitrary number of gradients
        through simple backsubstitution. All matrices need to be assembled only once.

    !!! info "Change in tooling"
        In the `DerivativeSolver`, we leave JAX and fall back to the `numpy`/`scipy` stack. While
        the sequential solver operation should not be mush slower on the CPU, we have to transfer
        the data back grom the offloading device. We plan to implement a GPU-compatible solver with
        [CuPy](https://cupy.dev/) in a future version, or in JAX as soon as it offers the necessary
        linear algebra tools.

    Methods:
        solve: Solve the linear system for the adjoint variable
    """

    # ----------------------------------------------------------------------------------------------
    def __init__(
        self,
        solution: jtFloat[jax.Array | npt.NDArray, "num_vertices"],
        sparse_partial_update_solution: spa.COO,
    ) -> None:
        r"""Constructor for the derivative solver.

        Initializes the causality-inspired permutation matrix $\mathbf{P}$, and afterwards the
        permuted system matrix $\bar{\mathbf{A}}$, which is triangular.

        Args:
            solution (jax.Array | npt.NDArray): Obtained solution of the Eikonal equation
            sparse_partial_update_solution (spa.COO): Matrix $\mathbf{G}_u$ in sparse COO format
        """
        num_points = solution.size
        solution = np.array(solution, dtype=np.float32)
        self._sparse_permutation_matrix = self._assemble_permutation_matrix(solution)
        self._sparse_system_matrix = self._assemble_system_matrix(
            sparse_partial_update_solution, num_points
        )

    # ----------------------------------------------------------------------------------------------
    def solve(
        self, right_hand_side: jtFloat[jax.Array | npt.NDArray, "num_vertices"]
    ) -> jtFloat[npt.NDArray, "num_parameters"]:
        r"""Solve the linear system for the parametric gradient.

        Following the notation from the class docstring,this method solves the linear system for
        the adjoint variable $\mathbf{v}$. Given a right-hand-side $\mathbf{l}_u$, this is a three-
        step process:

        1. Permute the right hand side $\bar{l}_u = \mathbf{P}l_u$
        2. Solve the linear system $\bar{\mathbf{A}}\bar{\mathbf{v}} = \bar{l}_u$
        3. Permute solution back to the original ordering
            $\mathbf{v} = \mathbf{P}^T\bar{\mathbf{v}}$

        Args:
            right_hand_side (jax.Array | npt.NDArray): RHS for the linear system solve

        Returns:
            np.ndarray: Solution of the linear system solve, corresponding to the adjoint in an
                optimization context.
        """
        right_hand_side = np.array(right_hand_side, dtype=np.float32)
        permutated_right_hand_side = self._sparse_permutation_matrix @ right_hand_side
        permutated_solution = sp.sparse.linalg.spsolve_triangular(
            self._sparse_system_matrix, permutated_right_hand_side, lower=False, unit_diagonal=True
        )
        solution = self._sparse_permutation_matrix.T @ permutated_solution

        return solution

    # ----------------------------------------------------------------------------------------------
    def _assemble_permutation_matrix(
        self, solution: jtFloat[npt.NDArray, "num_vertices"]
    ) -> sp.sparse.csc_matrix:
        r"""Construct permutation matrix $\mathbf{P}$ for index ordering."""
        num_points = solution.size
        permutation_row_inds = np.arange(solution.size)
        permutation_col_inds = np.argsort(solution)
        permutation_values = np.ones(solution.size)
        sparse_permutation_matrix = sp.sparse.csc_matrix(
            (permutation_values, (permutation_row_inds, permutation_col_inds)),
            shape=(num_points, num_points),
        )

        return sparse_permutation_matrix

    # ----------------------------------------------------------------------------------------------
    def _assemble_system_matrix(
        self,
        sparse_partial_update_solution: spa.COO,
        num_points: int,
    ) -> sp.sparse.csc_matrix:
        r"""Assemble system matrix $\bar{\mathbf{A}}$ for gradient solver.

        Before invoking this method, the permutation matrix $\mathbf{P}$ must be initialized.
        """
        sparse_partial_update_solution = sparse_partial_update_solution.tocsc()
        sparse_identity_matrix = sp.sparse.identity(num_points, format="csc")
        sparse_system_matrix = sparse_identity_matrix - sparse_partial_update_solution
        sparse_system_matrix = (
            self._sparse_permutation_matrix
            @ sparse_system_matrix.T
            @ self._sparse_permutation_matrix.T
        )

        return sparse_system_matrix

    # ----------------------------------------------------------------------------------------------
    @property
    def sparse_system_matrix(self) -> sp.sparse.csc_matrix:
        r"""Get system matrix $\bar{\mathbf{A}}\in\mathbb{R}^{{N_V}\times {N_V}}$."""
        return self._sparse_system_matrix

    # ----------------------------------------------------------------------------------------------
    @property
    def sparse_permutation_matrix(self) -> sp.sparse.csc_matrix:
        r"""Get permutation matrix $\mathbf{P}\in\mathbb{R}^{{N_V}\times {N_V}}$."""
        return self._sparse_permutation_matrix


# ==================================================================================================
def compute_eikonax_jacobian(
    derivative_solver: DerivativeSolver,
    partial_derivative_parameter: spa.SparseArray | sp.sparse.spmatrix,
) -> npt.NDArray:
    """Compute Jacobian from concatenation of gradients, computed with unit vector RHS.

    !!! warning

        This method should only be used for small problems.

    Args:
        derivative_solver (DerivativeSolver): Initialized derivative solver object
        partial_derivative_parameter (spa.SparseArray | sp.sparse.spmatrix):
            Partial derivative of the global update operator with respect to the parameter vector

    Returns:
        npt.NDArray: (Dense) Jacobian matrix
    """
    rhs_adjoint = np.zeros(derivative_solver.sparse_permutation_matrix.shape[0])
    jacobian = []

    for i, _ in enumerate(rhs_adjoint):
        rhs_adjoint[i] = 1.0
        adjoint = derivative_solver.solve(rhs_adjoint)
        rhs_adjoint[i] = 0.0
        jacobian_row = partial_derivative_parameter.T @ adjoint
        jacobian.append(jacobian_row)
    jacobian = np.vstack(jacobian)
    return jacobian


# --------------------------------------------------------------------------------------------------
def compute_eikonax_hessian() -> None:
    """Compute Hessian matrix.

    !!! failure "Not implemented yet"
    """
    raise NotImplementedError



"""Core functions for Eikonax forward solves and parametric derivatives.

This module contains atomic functions that make up the Eikonax solver routines. They (and their
automatic derivatives computed with JAX) are further used to evaluate parametric derivatives.

Functions:
    compute_softminmax: Smooth double ReLU-type approximation that restricts a variable to the
        interval [0, 1].
    compute_edges: Compute the edges of a triangle from vertex indices and coordinates.
    compute_optimal_update_parameters_soft: Compute position parameter for update of a node within a
        specific triangle.
    compute_optimal_update_parameters_hard: Compute position parameter for update of a node within a
        specific triangle.
    _compute_optimal_update_parameters: Compute the optimal update parameter for the solution of the
        Eikonal equation.
    compute_fixed_update: Compute update for a given vertex, triangle, and update parameter.
    compute_update_candidates_from_adjacent_simplex: Compute all possible update candidates from an
        adjacent triangle.
    compute_vertex_update_candidates: Compute all update candidates for a given vertex.
    grad_average: JAX-compatible computation of the gradient of the average function.
"""

from numbers import Real

import jax
import jax.numpy as jnp
from jaxtyping import Float as jtFloat
from jaxtyping import Int as jtInt
from jaxtyping import Real as jtReal


# ==================================================================================================
@jax.custom_gradient
def compute_softminmax(value, order):  # noqa: ANN001,ANN201
    r"""Smooth double ReLU-type approximation that restricts a variable to the interval $[0, 1]$.

    Given an input value $x$ and an order parameter $\kappa > 0$, the method performs a
    differentiable transformation according to

    $$
    \begin{gather*}
        f_{\text{lb}}(x) = \frac{1}{\kappa}\log\Big[ 1 + \exp\big( \kappa x \big) \Big], \\
        f_{\text{ub}}(x) = 1 - \frac{1}{\kappa}\log\Big[ 1 + \exp\big( -\kappa(x-1) \big) \Big], \\
        \tilde{\phi}(x) = f_{\text{ub}}(f_{\text{lb}}(x)), \\
        \phi_{\text{lb}} = 1 - \frac{\log\big(1+\exp(\kappa)\big)}{\kappa} < 0, \\
        \phi(x) = \frac{\tilde{\phi}(x) - \phi_{\text{lb}}}{1-\phi_{\text{lb}}}.
    \end{gather*}
    $$

    The method is numerically stable, obeys the value range, and does not introduce any new extrema.

    !!! note
        The function is not typed to avoid issues with the beartype type checker in combination
        with JAX's `custom_gradient` decorator.

    Args:
        value (jax.Array): variable to restrict to range [0,1]
        order (int): Approximation order of the smooth approximation

    Returns:
        jax.Array: Smoothed/restricted value
    """
    lower_bound = -jnp.log(1 + jnp.exp(-order)) / order
    soft_value = jnp.where(
        value <= 0,
        jnp.log(1 + jnp.exp(order * value)) / order,
        value + jnp.log(1 + jnp.exp(-order * value)) / order,
    )
    soft_value = jnp.where(
        soft_value <= 1,
        soft_value - jnp.log(1 + jnp.exp(order * (soft_value - 1))) / order,
        1 - jnp.log(1 + jnp.exp(-order * (soft_value - 1))) / order,
    )
    soft_value = (soft_value - lower_bound) / (1 - lower_bound)

    def _grad_softminmax(
        value_increment: jtReal[jax.Array, "..."],
    ) -> tuple[jtReal[jax.Array, "..."], None]:
        term_1 = 1 + jnp.exp(-order) + jnp.exp(order * (value - 1))
        term_2 = 1 + jnp.exp(-order * value)
        grad = 1 / (term_1 * term_2)
        grad = grad / (1 - lower_bound)
        return grad * value_increment, None

    return soft_value, _grad_softminmax


# --------------------------------------------------------------------------------------------------
def compute_edges(
    i: jtInt[jax.Array, ""],
    j: jtInt[jax.Array, ""],
    k: jtInt[jax.Array, ""],
    vertices: jtFloat[jax.Array, "num_vertices dim"],
) -> tuple[jtFloat[jax.Array, "dim"], jtFloat[jax.Array, "dim"], jtFloat[jax.Array, "dim"]]:
    """Compute the edges of a triangle from vertex indices and coordinates.

    Args:
        i (jax.Array): First vertex index of a triangle
        j (jax.Array): Second vertex index of a triangle
        k (jax.Array): Third vertex index of a triangle
        vertices (jax.Array): jax.Array of all vertex coordinates

    Returns:
        tuple[jax.Array, jax.Array, jax.Array]: Triangle edge vectors
    """
    e_ji = vertices[i] - vertices[j]
    e_ki = vertices[i] - vertices[k]
    e_jk = vertices[k] - vertices[j]
    return e_ji, e_ki, e_jk


# --------------------------------------------------------------------------------------------------
def compute_optimal_update_parameters_soft(
    solution_values: jtFloat[jax.Array, "2"],
    parameter_tensor: jtFloat[jax.Array, "dim dim"],
    edges: tuple[jtFloat[jax.Array, "dim"], jtFloat[jax.Array, "dim"], jtFloat[jax.Array, "dim"]],
    softminmax_order: int,
    softminmax_cutoff: Real,
) -> jtFloat[jax.Array, "4"]:
    """Compute position parameter for update of a node within a specific triangle.

    For a given vertex $i$ and adjacent triangle, we compute the update for the solution of the
    Eikonal as propagating from a point on the connecting edge of the opposite vertices $j$ and $k$.
    We thereby assume the solution value to vary linearly on that edge. The update parameter in
    $[0,1]$ indicates the optimal linear combination of the solution values at $j$ and $k$, in the
    sense that the solution value at $i$ is minimized. As the underlying optimization problem is
    constrained, we compute the solutions of the unconstrained problem, as well as the boundary
    values. The former are constrained to the feasible region [0,1] by the soft minmax function
    implemented in [`compute_softminmax`][eikonax.corefunctions.compute_softminmax].
    We further clip values lying to far outside the feasible region, by masking them with value -1.
    This function is a wrapper, for the unconstrained solution values, it calls the implementation
    function
    [`_compute_optimal_update_parameters`][eikonax.corefunctions._compute_optimal_update_parameters].

    Args:
        solution_values (jax.Array): Current solution values, as per the previous iteration
        parameter_tensor (jax.Array): Parameter tensor for the current triangle
        edges (tuple[jax.Array, jax.Array, jax.Array]): Edge vectors of the triangle under
            consideration
        softminmax_order (int): Order of the soft minmax function to be employed
        softminmax_cutoff (int): Cutoff value beyond parameter values are considered infeasible
            and masked with -1

    Returns:
        jax.Array: All possible candidates for the update parameter, according to the solution
            of the constrained optimization problem
    """
    lambda_1, lambda_2 = _compute_optimal_update_parameters(
        solution_values, parameter_tensor, edges
    )
    lambda_1_clipped = compute_softminmax(lambda_1, softminmax_order)
    lambda_2_clipped = compute_softminmax(lambda_2, softminmax_order)
    lower_bound = -softminmax_cutoff
    upper_bound = 1 + softminmax_cutoff

    lambda_1_clipped = jnp.where(
        (lambda_1 < lower_bound) | (lambda_1 > upper_bound), -1, lambda_1_clipped
    )
    lambda_2_clipped = jnp.where(
        (lambda_2 < lower_bound) | (lambda_2 > upper_bound) | (lambda_2 == lambda_1),
        -1,
        lambda_2_clipped,
    )
    lambda_array = jnp.array((0, 1, lambda_1_clipped, lambda_2_clipped), dtype=jnp.float32)
    return lambda_array


# --------------------------------------------------------------------------------------------------
def compute_optimal_update_parameters_hard(
    solution_values: jtFloat[jax.Array, "2"],
    parameter_tensor: jtFloat[jax.Array, "dim dim"],
    edges: tuple[jtFloat[jax.Array, "dim"], jtFloat[jax.Array, "dim"], jtFloat[jax.Array, "dim"]],
) -> jtFloat[jax.Array, "4"]:
    """Compute position parameter for update of a node within a specific triangle.

    For a given vertex $i$ and adjacent triangle, we compute the update for the solution of the
    Eikonal as propagating from a point on the connecting edge of the opposite vertices $j$ and $k$.
    We thereby assume the solution value to vary linearly on that dge. The update parameter in
    $[0,1]$ indicates the optimal linear combination of the solution values at $j$ and $k$, in the
    sense that the solution value at $i$ is minimized. As the underlying optimization problem is
    constrained, we compute the solutions of the unconstrained problem, as well as the boundary
    values. The former are constrained to the feasible region $[0,1]$ by a simple cutoff.
    This function is a wrapper, for the unconstrained solution values, it calls the implementation
    function
    [`_compute_optimal_update_parameters`][eikonax.corefunctions._compute_optimal_update_parameters].

    Args:
        solution_values (jax.Array): Current solution values, as per the previous iteration
        parameter_tensor (jax.Array): Parameter tensor for the current triangle
        edges (tuple[jax.Array, jax.Array, jax.Array]): Edge vectors of the triangle under
            consideration

    Returns:
        jax.Array: All possible candidates for the update parameter, according to the solution
            of the constrained optimization problem
    """
    lambda_1, lambda_2 = _compute_optimal_update_parameters(
        solution_values, parameter_tensor, edges
    )
    lambda_1_clipped = jnp.where((lambda_1 <= 0) | (lambda_1 >= 1), -1, lambda_1)
    lambda_2_clipped = jnp.where(
        (lambda_2 <= 0) | (lambda_2 >= 1) | (lambda_2 == lambda_1),
        -1,
        lambda_2,
    )
    lambda_array = jnp.array((0, 1, lambda_1_clipped, lambda_2_clipped), dtype=jnp.float32)
    return lambda_array


# --------------------------------------------------------------------------------------------------
def _compute_optimal_update_parameters(
    solution_values: jtFloat[jax.Array, "2"],
    parameter_tensor: jtFloat[jax.Array, "dim dim"],
    edges: tuple[jtFloat[jax.Array, "dim"], jtFloat[jax.Array, "dim"], jtFloat[jax.Array, "dim"]],
) -> tuple[jtFloat[jax.Array, ""], jtFloat[jax.Array, ""]]:
    r"""Compute the optimal update parameter for the solution of the Eikonal equation.

    The function works for the update within a given triangle. The solutions of the unconstrained
    minimization problem are given as the roots of a quadratic polynomial. For the local metric
    tensor $M_s$ for the given simplex, the set of known solution values $\mathbf{U}_{jk}$, and the
    simplex edges $\mathbf{E}_{ijk}$, we have that

    $$
        \lambda_{1/2}(\mathbf{M}_{s},\mathbf{U}_{jk},\mathbf{E}_{ijk}) =
        \frac{(\mathbf{e}_{jk},\mathbf{M}_{s}^{-1}\mathbf{e}_{ki})
        \pm c(\mathbf{M}_{s},\mathbf{U}_{jk},\mathbf{E}_{ijk})}{(\mathbf{e}_{jk},
        \mathbf{M}_{s}^{-1}\mathbf{e}_{jk})},
    $$

    with

    $$
        c(\mathbf{M}_{s},\mathbf{U}_{jk},\mathbf{E}_{ijk}) =
        (u_k - u_j) \sqrt{\frac{(\mathbf{e}_{jk},\mathbf{M}_{s}^{-1}\mathbf{e}_{jk})
        (\mathbf{e}_{ki},\mathbf{M}_{s}^{-1}\mathbf{e}_{ki})
        - (\mathbf{e}_{jk},\mathbf{M}_{s}^{-1}\mathbf{e}_{ki})^2}
        {(\mathbf{e}_{jk},\mathbf{M}_{s}^{-1}\mathbf{e}_{jk}) - (u_k - u_j)^2}}.
    $$

    These roots may or may not lie inside the feasible region $[0,1]$.The function returns both
    solutions, which are then further processed in the calling wrapper.
    """
    u_j, u_k = solution_values
    e_ji, _, e_jk = edges
    delta_u = u_k - u_j
    a_1 = jnp.dot(e_jk, parameter_tensor @ e_jk)
    a_2 = jnp.dot(e_jk, parameter_tensor @ e_ji)
    a_3 = jnp.dot(e_ji, parameter_tensor @ e_ji)

    nominator = a_1 * a_3 - a_2**2
    denominator = a_1 - delta_u**2
    # Treat imaginary roots as inf
    sqrt_term = jnp.where(denominator > 0, jnp.sqrt(nominator / denominator), jnp.inf)
    c = delta_u * sqrt_term

    lambda_1 = (a_2 + c) / a_1
    lambda_2 = (a_2 - c) / a_1
    return lambda_1, lambda_2


# --------------------------------------------------------------------------------------------------
def compute_fixed_update(
    solution_values: jtFloat[jax.Array, "2"],
    parameter_tensor: jtFloat[jax.Array, "dim dim"],
    lambda_value: jtFloat[jax.Array, ""],
    edges: tuple[jtFloat[jax.Array, "dim"], jtFloat[jax.Array, "dim"], jtFloat[jax.Array, "dim"]],
) -> jtFloat[jax.Array, ""]:
    r"""Compute update for a given vertex, triangle, and update parameter.

    The update value is given by the solution at a point  on the edge between the opposite vertices,
    plus the travel time from that point to the vertices under consideration. For a local metric
    tensor $M_s$ for the given simplex, the set of known solution values $\mathbf{U}_{jk}$, the
    simplex edges $\mathbf{E}_{ijk}$, and an optimal update parameter $\lambda$, it reads

    $$
        G_{\text{fix}}(\lambda,\mathbf{M}_s,\mathbf{U}_{jk}, \mathbf{E}_{ijk}) =
        u_j + \lambda (u_k-u_j) + \sqrt{\big((\mathbf{e}_{ji} - \lambda\mathbf{e}_{jk}),
        \mathbf{M}_{s}^{-1}(\mathbf{e}_{ji} - \lambda\mathbf{e}_{jk})\big)}
    $$

    Args:
        solution_values (jax.Array): Current solution values at opposite vertices j and k,
            as per the previous iteration
        parameter_tensor (jax.Array): Conductivity tensor for the current triangle
        lambda_value (jax.Array): Optimal update parameter
        edges (tuple[jax.Array, jax.Array, jax.Array]): Edge vectors of the triangle under
            consideration

    Returns:
        jax.Array: Updated solution value for the vertex under consideration
    """
    u_j, u_k = solution_values
    e_ji, _, e_jk = edges
    diff_vector = e_ji - lambda_value * e_jk
    transport_term = jnp.sqrt(jnp.dot(diff_vector, parameter_tensor @ diff_vector))
    update = lambda_value * u_k + (1 - lambda_value) * u_j + transport_term
    return update


# --------------------------------------------------------------------------------------------------
def compute_update_candidates_from_adjacent_simplex(
    old_solution_vector: jtFloat[jax.Array, "num_vertices"],
    tensor_field: jtFloat[jax.Array, "num_simplices dim dim"],
    adjacency_data: jtInt[jax.Array, "max_num_adjacent_simplices"],
    vertices: jtFloat[jax.Array, "num_vertices dim"],
    use_soft_update: bool,
    softminmax_order: int | None,
    softminmax_cutoff: Real | None,
) -> tuple[jtFloat[jax.Array, "4"], jtFloat[jax.Array, "4"]]:
    """Compute all possible update candidates from an adjacent triangle.

    Given a vertex and an adjacent triangle, this method computes all optimal update parameter
    candidates and the corresponding update values. To obey JAX's homogeneous array requirement,
    update values are also computed for infeasible parameter values, and have to be masked in the
    calling routine. This methods basically collects all results from the
    [`compute_optimal_update_parameters`][eikonax.corefunctions.compute_optimal_update_parameters_soft]
    and [`compute_fixed_update`][eikonax.corefunctions.compute_fixed_update] functions.

    Args:
        old_solution_vector (jax.Array): Given solution vector, as per a previous iteration
        tensor_field (jax.Array): Array of all tensor fields
        adjacency_data (jax.Array): Index of one adjaccent triangle and corresponding vertices
        vertices (jax.Array): Array of all vertex coordinates
        use_soft_update (bool): Flag to indicate whether to use a soft update or a hard update
        softminmax_order (int | None): Order of the soft minmax function for the update parameter,
            see `compute_softminmax`. Only required for `use_soft_update=True`
        softminmax_cutoff (Real | None): Cutoff value for the soft minmax computation, see
            `compute_optimal_update_parameters_soft`. Only required for `use_soft_update=True`

    Returns:
        tuple[jax.Array, jax.Array]: Update values and parameter candidates from the given
            triangle
    """
    i, j, k, s = adjacency_data
    solution_values = jnp.array((old_solution_vector[j], old_solution_vector[k]), dtype=jnp.float32)
    edges = compute_edges(i, j, k, vertices)
    parameter_tensor = tensor_field[s]
    if use_soft_update:
        lambda_array = compute_optimal_update_parameters_soft(
            solution_values, parameter_tensor, edges, softminmax_order, softminmax_cutoff
        )
    else:
        lambda_array = compute_optimal_update_parameters_hard(
            solution_values, parameter_tensor, edges
        )
    update_candidates = jnp.zeros(4)

    for i, lambda_candidate in enumerate(lambda_array):
        update = compute_fixed_update(solution_values, parameter_tensor, lambda_candidate, edges)
        update_candidates = update_candidates.at[i].set(update)
    return update_candidates, lambda_array


# --------------------------------------------------------------------------------------------------
def compute_vertex_update_candidates(
    old_solution_vector: jtFloat[jax.Array, "num_vertices"],
    tensor_field: jtFloat[jax.Array, "num_simplices dim dim"],
    adjacency_data: jtInt[jax.Array, "max_num_adjacent_simplices 4"],
    vertices: jtFloat[jax.Array, "num_vertices dim"],
    use_soft_update: bool,
    softminmax_order: int,
    softminmax_cutoff: Real,
) -> jtFloat[jax.Array, "max_num_adjacent_simplices 4"]:
    """Compute all update candidates for a given vertex.

    This method combines all updates from adjacent triangles to a given vertex, as computed in the
    function
    [`compute_update_candidates_from_adjacent_simplex`][eikonax.corefunctions.compute_update_candidates_from_adjacent_simplex].
    Infeasible candidates are masked with `jnp.inf`.

    Args:
        old_solution_vector (jax.Array): Given solution vector, as per a previous iteration
        tensor_field (jax.Array): jax.Array of all tensor fields
        adjacency_data (jax.Array): Data of all adjacent triangles and corresponding vertices
        vertices (jax.Array): jax.Array of all vertex coordinates
        use_soft_update (bool): Flag to indicate whether to use a soft update or a hard update
        softminmax_order (int | None): Order of the soft minmax function for the update parameter,
            see `compute_softminmax`. Only required for `use_soft_update=True`
        softminmax_cutoff (Real | None): Cutoff value for the soft minmax computation, see
            `compute_optimal_update_parameters_soft`. Only required for `use_soft_update=True`

    Returns:
        jax.Array: All possible update values for the given vertex, infeasible vertices are masked
            with jnp.inf
    """
    max_num_adjacent_simplices = adjacency_data.shape[0]
    vertex_update_candidates = jnp.zeros((max_num_adjacent_simplices, 4), dtype=jnp.float32)
    lambda_arrays = jnp.zeros((max_num_adjacent_simplices, 4), dtype=jnp.float32)

    for i, indices in enumerate(adjacency_data):
        simplex_update_candidates, lambda_array_candidates = (
            compute_update_candidates_from_adjacent_simplex(
                old_solution_vector,
                tensor_field,
                indices,
                vertices,
                use_soft_update,
                softminmax_order,
                softminmax_cutoff,
            )
        )
        vertex_update_candidates = vertex_update_candidates.at[i, :].set(simplex_update_candidates)
        lambda_arrays = lambda_arrays.at[i, :].set(lambda_array_candidates)

    # Mask infeasible updates
    # 1. Not an adjacent triangle, only buffer/fill value in adjacency data
    # 2. Infeasible update parameter, indicated with -1
    active_simplex_inds = adjacency_data[:, 3]
    vertex_update_candidates = jnp.where(
        (active_simplex_inds[..., None] != -1) & (lambda_arrays != -1),
        vertex_update_candidates,
        jnp.inf,
    )
    return vertex_update_candidates


# ==================================================================================================
"""Derivatives of elementary function based on JAX's AD capabilities."""
# Derivative of update value function w.r.t. current solution values, 1x2
grad_update_solution = jax.grad(compute_fixed_update, argnums=0)
# Deritative of update value function w.r.t. parameter tensor, 1xDxD
grad_update_parameter = jax.grad(compute_fixed_update, argnums=1)
# Derivative of update value function w.r.t. update parameter, 1x1
grad_update_lambda = jax.grad(compute_fixed_update, argnums=2)
# Derivative of update parameter function w.r.t. current solution values, 2x2
jac_lambda_soft_solution = jax.jacobian(compute_optimal_update_parameters_soft, argnums=0)
jac_lambda_hard_solution = jax.jacobian(compute_optimal_update_parameters_hard, argnums=0)
# Derivative of update parameter function w.r.t. parameter tensor, 2xDxD
jac_lambda_soft_parameter = jax.jacobian(compute_optimal_update_parameters_soft, argnums=1)
jac_lambda_hard_parameter = jax.jacobian(compute_optimal_update_parameters_hard, argnums=1)


# --------------------------------------------------------------------------------------------------
def grad_average(
    args: jtFloat[jax.Array, "num_args"], min_arg: jtFloat[jax.Array, ""]
) -> jtFloat[jax.Array, "num_args"]:
    """JAX-compatible computation of the gradient of the average function.

    This function is applied to actual minimum values, meaning it does not have a purpose
    on the evaluation level. It renders the minimum computation differentiable, however.
    Specifically, consider the scenario where a vertex can be updated identically from two different
    directions. Clearly, the mapping from the parameter field to the solution vector is not
    continuously differentiable at this point. Instead, we can only formulate a possible set
    of sub-gradients. One strategy to cope with sub-gradients is to employ simple averaging over
    all possible candidates.
    """
    num_min_args = jnp.count_nonzero(args == min_arg)
    average_grad = jnp.where(args == min_arg, 1 / num_min_args, 0)
    return average_grad



r"""Composable and differentiable parameter tensor fields.

This module provides ABCs and implementations for the creation of differentiable parameter fields
used in Eikonax. Recall that for the Eikonax solver, and particularly parameteric derivatives, we
require an input tensor field
$\mathbf{M}: \mathbb{R}^M \times \mathbb{N}_0 \to \mathbb{S}_+^{d\times d}$. This means that the
tensor field is a mapping $\mathbf{M}(\mathbf{m},s)$ that assigns, given a global parameter vector
$\mathbf{m}$, an s.p.d tensor to every simplex $s$ in the mesh. To allow for sufficient flexibility
in the choice of tensor field, we implement it as a composition of two main components.

1. [`AbstractVectorToSimplicesMap`][eikonax.tensorfield.AbstractVectorToSimplicesMap] provides the
    interface for a mapping from the global parameter vector $\mathbf{m}$ to the local parameter
    values $\mathbf{m}_s$ required to assemble the tensor $\mathbf{M}_s$ for simplex $s$.
2. [`AbstractSimplexTensor`][eikonax.tensorfield.AbstractSimplexTensor] provides the interface for
    the assembly of the local tensor $\mathbf{M}_s$, given the local contributions $\mathbf{m}_s$
    and a simplex s.

Concrete implementations of both components are used to initialize the
[`TensorField`][eikonax.tensorfield.TensorField] object, which vectorizes and differentiates them
using JAX, to provide the mapping $\mathbf{M}(\mathbf{m})$ and its Jacobian tensor
$\frac{d \mathbf{M}}{d \mathbf{m}}$.

Classes:
    AbstractVectorToSimplicesMap: ABC interface contract for vector-to-simplices maps
    LinearScalarMap: Simple one-to-one map from global to simplex parameters
    AbstractSimplexTensor: ABC interface contract for assembly of the tensor field
    LinearScalarSimplexTensor: SimplexTensor implementation relying on one parameter per simplex
    InvLinearScalarSimplexTensor: SimplexTensor implementation relying on one parameter per simplex
    TensorField: Tensor field component
"""

from abc import abstractmethod
from typing import final

import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
import numpy.typing as npt
import sparse as spa
from jaxtyping import Float as jtFloat
from jaxtyping import Int as jtInt
from jaxtyping import Real as jtReal


# ==================================================================================================
class AbstractVectorToSimplicesMap(eqx.Module):
    """ABC interface contract for vector-to-simplices maps.

    Every component derived from this class needs to implement the `map`  and `derivative`  methods.
    The `map` method is responsible for returning the relevant parameters for a given simplex from
    the global parameter vector. The `derivative` method computes the derivative of the mapping with
    respect to the global parameters.

    Methods:
        map: Interface for vector-to-simplex mapping
        derivative: Interface for vector-to-simplex mapping derivative
    """

    # ----------------------------------------------------------------------------------------------
    @abstractmethod
    def map(
        self, simplex_ind: jtInt[jax.Array, ""], parameters: jtReal[jax.Array, "num_parameters"]
    ) -> jtReal[jax.Array, "num_parameters_local"]:
        """Interface for vector-to-simplex mapping.

        For the given `simplex_ind`, return those parameters from the global parameter vector that
        are relevant for the simplex. This methods need to be broadcastable over `simplex_ind` by
        JAX (with `vmap`).

        Args:
            simplex_ind (int): Index of the simplex under consideration
            parameters (jax.Array): Global parameter vector

        Raises:
            NotImplementedError: ABC error indicating that the method needs to be implemented
                in subclasses

        Returns:
            jax.Array: Relevant parameters for the simplex
        """
        raise NotImplementedError

    # ----------------------------------------------------------------------------------------------
    @abstractmethod
    def derivative(
        self, simplex_ind: jtInt[jax.Array, ""], parameters: jtReal[jax.Array, "num_parameters"]
    ) -> tuple[
        jtReal[jax.Array, "num_parameters_local"],
        jtReal[jax.Array, "num_parameters_local"],
    ]:
        r"""Interface for vector-to-simplex mapping derivative.

        The derivative of the simplex mapping for a given `simplex_ind` is a `num_local_parameters`
        $\times$ `num_parameters` Jacobian matrix, where the only non-zero entries are in the
        columns corresponding to the values in the global parameter that are used for the current
        simplex. The corresponding entries are typically identical to one. This method needs to
        return the Jacobian matrix in sparse COO format, i.e. one vector comprising the column
        numbers, and one vector comprising the matrix entries.

        Args:
            simplex_ind (int): Index of the simplex under consideration
            parameters (jax.Array): Global parameter vector

        Raises:
            NotImplementedError: ABC error indicating that the method needs to be implemented
                in subclasses

        Returns:
            tuple[jax.Array, jax.Array]: Jacobi matrix, expressed via relevant global indices and
                matrix entries
        """
        raise NotImplementedError


# --------------------------------------------------------------------------------------------------
@final
class LinearScalarMap(AbstractVectorToSimplicesMap):
    r"""Simple one-to-one map from global to simplex parameters.

    Every simplex takes exactly one parameter $m_s$, which is sorted in the global parameter
    in the same order as the simplices, meaning that $m_s = \mathbf{m}[s]$.
    """

    # ----------------------------------------------------------------------------------------------
    def map(
        self,
        simplex_ind: jtInt[jax.Array, ""],
        parameters: jtReal[jax.Array, "num_parameters"],
    ) -> jtReal[jax.Array, "num_parameters_local"]:
        """Return relevant parameters for a given simplex.

        Args:
            simplex_ind (int): Index of the simplex under consideration
            parameters (jax.Array): Global parameter vector

        Returns:
            jax.Array: relevant parameter (only one)
        """
        parameter = jnp.expand_dims(parameters[simplex_ind], axis=-1)
        return parameter

    # ----------------------------------------------------------------------------------------------
    def derivative(
        self, simplex_ind: jtInt[jax.Array, ""], _parameters: jtReal[jax.Array, "num_parameters"]
    ) -> tuple[
        jtReal[jax.Array, "num_parameters_local"],
        jtReal[jax.Array, "num_parameters_local"],
    ]:
        """Return sparse representation of Jacobi matrix, in this case two arrays of size one.

        Args:
            simplex_ind (int): Index of the simplex under consideration
            _parameters (jax.Array): Global parameter vector (not used)

        Returns:
            tuple[jax.Array, jax.Array]: Jacobi matrix, expressed via relevant global indices and
                matrix entries
        """
        global_parameter_inds = jnp.array((simplex_ind,), dtype=jnp.int32)
        derivative_vals = jnp.ones_like(global_parameter_inds, dtype=jnp.float32)
        return global_parameter_inds, derivative_vals


# ==================================================================================================
class AbstractSimplexTensor(eqx.Module):
    """ABC interface contract for assembly of the tensor field.

    `SimplexTensor` components assemble the tensor field for a given simplex and a set of parameters
    for that simplex. The relevant parameters are provided by the `VectorToSimplicesMap` component
    from the global parameter vector.

    !!! note
        This class provides the metric tensor as used in the inner product for the update stencil of
        the eikonal equation. This is the **inverse** of the conductivity tensor, which is the
        actual tensor field in the eikonal equation.

    Methods:
        assemble: Assemble the tensor field for a given simplex and parameters
        derivative: Parametric derivative of the `assemble` method
    """

    # Equinox modules are data classes, so we have to define attributes at the class level
    dimension: int

    # ----------------------------------------------------------------------------------------------
    @abstractmethod
    def assemble(
        self,
        simplex_ind: jtInt[jax.Array, ""],
        parameters: jtFloat[jax.Array, "num_parameters_local"],
    ) -> jtFloat[jax.Array, "dim dim"]:
        r"""Assemble the tensor field for given simplex and parameters.

        Given a parameter array of size $m_s$, the method returns a tensor of size $d\times d$.
        The method needs to be broadcastable over `simplex_ind` by JAX (with `vmap`).

        Args:
            simplex_ind (int): Index of the simplex under consideration
            parameters (jax.Array): Parameters for the simplex

        Raises:
            NotImplementedError: ABC error indicating that the method needs to be implemented
                in subclasses

        Returns:
            jax.Array: Tensor field for the simplex under consideration
        """
        raise NotImplementedError

    # ----------------------------------------------------------------------------------------------
    @abstractmethod
    def derivative(
        self,
        simplex_ind: jtInt[jax.Array, ""],
        parameters: jtFloat[jax.Array, "num_parameters_local"],
    ) -> jtFloat[jax.Array, "dim dim num_local_parameters"]:
        r"""Parametric derivative of the `assemble` method.

        Given a parameter array of size $m_s$, the methods returns a Jacobian tensor of size
        $d\times d\times m_s$. The method needs to be broadcastable over `simplex_ind` by JAX
        (with `vmap`).

        Args:
            simplex_ind (int): Index of the simplex under consideration
            parameters (jax.Array): Parameters for the simplex

        Raises:
            NotImplementedError: ABC error indicating that the method needs to be implemented
                in subclasses

        Returns:
            jax.Array: Jacobian tensor for the simplex under consideration
        """
        raise NotImplementedError


# ==================================================================================================
@final
class LinearScalarSimplexTensor(AbstractSimplexTensor):
    r"""SimplexTensor implementation relying on one parameter per simplex.

    Given a scalar parameter $m_s$, the tensor field is assembled as $m_s \cdot \mathbf{I}$, where
    $\mathbf{I}$ is the identity matrix.

    Methods:
        assemble: Assemble the tensor field for a parameter vector
        derivative: Parametric derivative of the `assemble` method
    """

    # ----------------------------------------------------------------------------------------------
    def assemble(
        self,
        _simplex_ind: jtInt[jax.Array, ""],
        parameters: jtFloat[jax.Array, "num_parameters_local"],
    ) -> jtFloat[jax.Array, "dim dim"]:
        """Assemble tensor for given simplex.

        the `parameters` argument is a scalar here, and `_simplex_ind` is not used.

        Args:
            _simplex_ind (int): Index of simplex under consideration (not used)
            parameters (jax.Array): Parameter (scalar) for tensor assembly

        Returns:
            jax.Array: Tensor for the simplex
        """
        tensor = parameters * jnp.identity(self.dimension, dtype=jnp.float32)
        return tensor

    # ----------------------------------------------------------------------------------------------
    def derivative(
        self,
        _simplex_ind: jtInt[jax.Array, ""],
        _parameters: jtFloat[jax.Array, "num_parameters_local"],
    ) -> jtFloat[jax.Array, "dim dim num_local_parameters"]:
        """Parametric derivative of the `assemble` method.

        Args:
            _simplex_ind (int): Index of simplex under consideration (not used)
            _parameters (jax.Array): Parameter (scalar) for tensor assembly

        Returns:
            jax.Array: Jacobian tensor for the simplex under consideration
        """
        derivative = jnp.expand_dims(jnp.identity(self.dimension, dtype=jnp.float32), axis=-1)
        return derivative


# ==================================================================================================
@final
class InvLinearScalarSimplexTensor(AbstractSimplexTensor):
    r"""SimplexTensor implementation relying on one parameter per simplex.

    Given a scalar parameter $m_s$, the tensor field is assembled as
    $\frac{1}{m_s} \cdot \mathbf{I}$, where $\mathbf{I}$ is the identity matrix.

    Methods:
        assemble: Assemble the tensor field for a parameter vector
        derivative: Parametric derivative of the `assemble` method
    """

    # ----------------------------------------------------------------------------------------------
    def assemble(
        self,
        _simplex_ind: jtInt[jax.Array, ""],
        parameters: jtFloat[jax.Array, "num_parameters_local"],
    ) -> jtFloat[jax.Array, "dim dim"]:
        """Assemble tensor for given simplex.

        The `parameters` argument is a scalar here, and `_simplex_ind` is not used.

        Args:
            _simplex_ind (int): Index of simplex under consideration (not used)
            parameters (jax.Array): Parameter (scalar) for tensor assembly

        Returns:
            jax.Array: Tensor for the simplex
        """
        tensor = 1 / parameters * jnp.identity(self.dimension, dtype=jnp.float32)
        return tensor

    # ----------------------------------------------------------------------------------------------
    def derivative(
        self,
        _simplex_ind: jtInt[jax.Array, ""],
        parameters: jtFloat[jax.Array, "num_local_parameters"],
    ) -> jtFloat[jax.Array, "dim dim num_local_parameters"]:
        """Parametric derivative of the `assemble` method.

        Args:
            _simplex_ind (int): Index of simplex under consideration (not used)
            parameters (jax.Array): Parameter (scalar) for tensor assembly

        Returns:
            jax.Array: Jacobian tensor for the simplex under consideration
        """
        derivative = (
            -1
            / jnp.square(parameters)
            * jnp.expand_dims(jnp.identity(self.dimension, dtype=jnp.float32), axis=-1)
        )
        return derivative


# ==================================================================================================
class TensorField(eqx.Module):
    r"""Tensor field component.

    Tensor fields combine the functionality of vector-to-simplices maps and simplex tensors
    according to the composition over inheritance principle. They constitute the full mapping
    $\mathbf{M}(\mathbf{m})$ from the global parameter vector to the tensor field over all mesh
    faces (simplices). In addition, they provide the parametric derivative
    $\frac{d\mathbf{M}}{\mathbf{m}}$ of that mapping. Tensor fields are completely independent from
    the Eikonax solver and derivator, but the output of these two components can be used to compute
    the partial derivative $\mathbf{G}_m = \frac{du}{d\mathbf{M}}\frac{d\mathbf{M}}{\mathbf{m}}$.

    Methods:
        assemble_field: Assemble the tensor field for the given parameter vector
        assemble_jacobian: Assemble the parametric derivative $\frac{d\mathbf{M}}{\mathbf{m}}$
            of the tensor field
    """

    # Equinox modules are data classes, so we have to define attributes at the class level
    _num_simplices: int
    _simplex_inds: jtInt[jax.Array, "num_simplices"]
    _vector_to_simplices_map: AbstractVectorToSimplicesMap
    _simplex_tensor: AbstractSimplexTensor

    # ----------------------------------------------------------------------------------------------
    def __init__(
        self,
        num_simplices: int,
        vector_to_simplices_map: AbstractVectorToSimplicesMap,
        simplex_tensor: AbstractSimplexTensor,
    ) -> None:
        """Constructor.

        Takes information about the mesh simplices, a vector-to-simplices map, and a simplex tensor
        map.

        Args:
            num_simplices (int): Number of simplices in the mesh
            vector_to_simplices_map (AbstractVectorToSimplicesMap): Mapping from global to simplex
                parameters
            simplex_tensor (AbstractSimplexTensor): Tensor field assembly for a given simplex
        """
        self._num_simplices = num_simplices
        self._simplex_inds = jnp.arange(num_simplices, dtype=jnp.int32)
        self._vector_to_simplices_map = vector_to_simplices_map
        self._simplex_tensor = simplex_tensor

    # ----------------------------------------------------------------------------------------------
    @eqx.filter_jit
    def assemble_field(
        self, parameter_vector: jtFloat[jax.Array | npt.NDArray, "num_parameters"]
    ) -> jtFloat[jax.Array, "num_simplices dim dim"]:
        """Assemble global tensor field from global parameter vector.

        This method simply chains calls to the vector-to-simplices map and the simplex tensor
        objects, vectorized over all simplices.

        Args:
            parameter_vector (jax.Array | npt.NDArray): Global parameter vector

        Returns:
            jax.Array: Global tensor field
        """
        parameter_vector = jnp.array(parameter_vector, dtype=jnp.float32)
        vector_to_simplices_global = jax.vmap(self._vector_to_simplices_map.map, in_axes=(0, None))
        simplex_tensor_assembly_global = jax.vmap(self._simplex_tensor.assemble, in_axes=(0, 0))
        simplex_params = vector_to_simplices_global(self._simplex_inds, parameter_vector)
        tensor_field = simplex_tensor_assembly_global(self._simplex_inds, simplex_params)

        return tensor_field

    # ----------------------------------------------------------------------------------------------
    def assemble_jacobian(
        self,
        parameter_vector: jtFloat[jax.Array | npt.NDArray, "num_parameters"],
    ) -> spa.COO:
        r"""Assemble the Jacobian $\frac{d\mathbf{M}}{d\mathbf{m}}$.

        The assembly of the Jacobian matrix works via a local chaining of the `derivative` calls
        to the simplex tensor and the vector-to-simplices map. A sparse COO representation of the
        call output is then vectorized over all simplices with JAX's `vmap`.

        Args:
            parameter_vector (jax.Array | npt.NDArray): Global parameter vector

        Returns:
            spa.COO: Jacobian tensor in sparse COO format,
                has dimension $N_S \times d \times d \times M$
        """
        parameter_vector = jnp.array(parameter_vector)
        coords, values = self._assemble_jacobian_global(parameter_vector)
        coords = [np.array(co.flatten(), dtype=np.int32) for co in coords]
        values = np.array(values.flatten(), dtype=np.float32)
        jacobian = spa.COO(
            coords,
            values,
            shape=(
                self._num_simplices,
                self._simplex_tensor.dimension,
                self._simplex_tensor.dimension,
                parameter_vector.size,
            ),
        )

        return jacobian

    # ----------------------------------------------------------------------------------------------
    @eqx.filter_jit
    def _assemble_jacobian_global(
        self, parameter_vector: jtFloat[jax.Array, "num_parameters"]
    ) -> tuple[
        list[jtInt[jax.Array, "num_matrix_entries"]], jtFloat[jax.Array, "num_matrix_entries"]
    ]:
        """Intermediate call for vectorization with `vmap` in `jit` context."""
        assemble_jacobian_global = jax.vmap(self._assemble_jacobian_local, in_axes=(0, None))
        *coords, values = assemble_jacobian_global(self._simplex_inds, parameter_vector)

        return coords, values

    # ----------------------------------------------------------------------------------------------
    def _assemble_jacobian_local(
        self,
        simplex_ind: jtInt[jax.Array, ""],
        parameter_vector: jtFloat[jax.Array, "num_parameters"],
    ) -> tuple[
        jtInt[jax.Array, "num_matrix_entries"],
        jtInt[jax.Array, "num_matrix_entries"],
        jtInt[jax.Array, "num_matrix_entries"],
        jtInt[jax.Array, "num_matrix_entries"],
        jtFloat[jax.Array, "derivative_dim_1 derivative_dim_2"],
    ]:
        """Assembly of sparse jacobian representation for single simplex."""
        local_parameters = self._vector_to_simplices_map.map(simplex_ind, parameter_vector)
        partial_derivative_simplex = self._simplex_tensor.derivative(simplex_ind, local_parameters)
        global_parameter_inds, map_derivative_values = self._vector_to_simplices_map.derivative(
            simplex_ind, parameter_vector
        )
        tensor_dim = self._simplex_tensor.dimension
        num_tensor_entries = tensor_dim**2
        num_global_parameters = global_parameter_inds.size

        simplex_inds = simplex_ind * jnp.ones(
            (num_tensor_entries * num_global_parameters,), dtype=jnp.int32
        )
        tensor_inds = jnp.arange(self._simplex_tensor.dimension)
        tensor_d1_inds = jnp.tile(jnp.repeat(tensor_inds, tensor_dim), num_global_parameters)
        tensor_d2_inds = jnp.tile(jnp.tile(tensor_inds, tensor_dim), num_global_parameters)
        parameters_inds = jnp.repeat(global_parameter_inds, num_tensor_entries)
        partial_derivative_values = partial_derivative_simplex @ map_derivative_values

        return (
            simplex_inds,
            tensor_d1_inds,
            tensor_d2_inds,
            parameters_inds,
            partial_derivative_values,
        )



"""Eikonax forward solver.

Classes:
    SolverData: Settings for the initialization of the Eikonax Solver.
    Solution: Eikonax solution object, returned by the solver.
    Solver: Eikonax solver class.
"""

import time
from dataclasses import dataclass
from numbers import Real
from typing import Annotated

import equinox as eqx
import jax
import jax.numpy as jnp
import numpy.typing as npt
from beartype.vale import Is, IsEqual
from jaxtyping import Bool as jtBool
from jaxtyping import Float as jtFloat
from jaxtyping import Int as jtInt

from . import corefunctions, logging, preprocessing


# ==================================================================================================
@dataclass
class SolverData:
    """Settings for the initialization of the Eikonax Solver.

    See the [Forward Solver](../usage/solve.md) documentation for more detailed explanations.

    Args:
        loop_type (str): Type of loop for iterations,
            options are 'jitted_for', 'jitted_while', 'nonjitted_while'.
        max_value (Real): Maximum value for the initialization of the solution vector.
        use_soft_update: Flag for using soft minmax approximation for optimization parameters
        softminmax_order (int | None): Order of the soft minmax approximation for optimization
            parameters. Only required if `use_soft_update` is True.
        softminmax_cutoff (Real | None): Cutoff distance from [0,1] for the soft minmax function.
            Only required if `use_soft_update` is True.
        max_num_iterations (int): Maximum number of iterations after which to terminate the solver.
            Required for all loop types
        tolerance (Real): Absolute difference between iterates in supremum norm, after which to
            terminate solver. Required for while loop types
        log_interval (int): Iteration interval after which log info is written. Required for
            non-jitted while loop type.
    """

    loop_type: Annotated[
        str, IsEqual["jitted_for"] | IsEqual["jitted_while"] | IsEqual["nonjitted_while"]
    ]
    max_value: Annotated[Real, Is[lambda x: x > 0]]
    use_soft_update: bool
    softminmax_order: Annotated[int, Is[lambda x: x > 0]] | None
    softminmax_cutoff: Annotated[Real, Is[lambda x: x > 0]] | None
    max_num_iterations: Annotated[int, Is[lambda x: x > 0]]
    tolerance: Annotated[Real, Is[lambda x: x > 0]] | None = None
    log_interval: Annotated[int, Is[lambda x: x > 0]] | None = None


@dataclass
class Solution:
    """Eikonax solution object, returned by the solver.

    See the [Forward Solver](../usage/solve.md) documentation for more detailed explanations.

    Args:
        values (jax.Array): Actual solution vector.
        num_iterations (int): Number of iterations performed in the solve.
        tolerance (float | jax.Array): Tolerance from last two iterates, or entire tolerance history
    """

    values: jtFloat[jax.Array, "num_vertices"]
    num_iterations: int | jtInt[jax.Array, ""]
    tolerance: float | jtFloat[jax.Array, "..."] | None = None


# ==================================================================================================
class Solver(eqx.Module):
    r"""Eikonax solver class.

    The solver class is the main component for computing the solution $u$ of the Eikonal equation
    for given geometry $\Omega$ of dimension $d$, tensor field $\mathbf{M}$, and initial sites
    $\Gamma$,

    $$
    \begin{gather*}
    \sqrt{\big(\nabla u(\mathbf{x}),\mathbf{M}(\mathbf{x})\nabla u(\mathbf{x})\big)} = 1,\quad
    \mathbf{x}\in\Omega, \\
    \nabla u(\mathbf{x}) \cdot \mathbf{n}(\mathbf{x}) \geq 0,\quad \mathbf{x}\in\partial\Omega, \\
    u(\mathbf{x}_0) = u_0,\quad \mathbf{x}_0 \in \Gamma.
    \end{gather*}
    $$

    On the discrete level, we assume that the eikonal equation is solved on a triangulation formed
    by $N_V$ vertices and $N_S$ associated triangles. This means that for a tensor field
    $\mathbf{M}\in\mathbb{R}^{N_S\times d\times d}$, the solver computes a solution vector
    $\mathbf{u}\in\mathbb{R}^{N_V}$. through iteraive updates

    $$
        \mathbf{u}^{(j+1)} = \mathbf{G}(\mathbf{u}^{(j)}, \mathbf{M}),
    $$

    where global update function $\mathbf{G}$ is derived from Godunov-type upwinding
    principles. The solver can either be run with a fixed number of iterations, or until a
    user-defined tolerance for the difference between two consecutive iterates in supremum norm is
    undercut.

    The Eikonax solver works on the vertex level,
    meaning that it considers updates from all adjacent triangles to a vertex, instead of all
    updates for all vertices per triangle. This allows to establish causality in the final solution,
    which is important for the efficient computation of parametric derivatives.
    The solver class is mainly a wrapper around different loop constructs, which call vectorized
    forms of the methods implemented in the [`corefunctions`][eikonax.corefunctions] module. These
    loop constructs evolve around the loop functionality provided by JAX.

    Methods:
        run: Main interface for Eikonax runs.
    """

    # Equinox modules are data classes, so specify attributes on class level
    _vertices: jax.Array
    _adjacency_data: jax.Array
    _loop_type: str
    _max_value: Real
    _use_soft_update: bool
    _softminmax_order: int | None
    _softminmax_cutoff: Real | None
    _max_num_iterations: int
    _initial_site_inds: jax.Array
    _initial_site_values: jax.Array
    _tolerance: Real | None
    _log_interval: int | None
    _logger: logging.Logger | None

    # ----------------------------------------------------------------------------------------------
    def __init__(
        self,
        mesh_data: preprocessing.MeshData,
        solver_data: SolverData,
        initial_sites: preprocessing.InitialSites,
        logger: logging.Logger | None = None,
    ) -> None:
        """Constructor of the solver class.

        The constructor initializes all data structures that are re-used in many-query scenarios,
        such as the solution of inverse problems.

        Args:
            mesh_data (preprocessing.MeshData): Vertex-based mesh data.
            solver_data (SolverData): Settings for the solver.
            initial_sites (preprocessing.InitialSites): vertices and values for source points
            logger (logging.Logger | None, optional): Logger object, only required for non-jitted
                while loops. Defaults to None.
        """
        self._vertices = mesh_data.vertices
        self._adjacency_data = mesh_data.adjacency_data
        self._loop_type = solver_data.loop_type
        self._max_value = solver_data.max_value
        self._use_soft_update = solver_data.use_soft_update
        self._softminmax_order = solver_data.softminmax_order
        self._softminmax_cutoff = solver_data.softminmax_cutoff
        self._max_num_iterations = solver_data.max_num_iterations
        self._tolerance = solver_data.tolerance
        self._log_interval = solver_data.log_interval
        self._initial_site_inds = initial_sites.inds
        self._initial_site_values = initial_sites.values
        self._logger = logger

    # ----------------------------------------------------------------------------------------------
    def run(
        self,
        tensor_field: jtFloat[jax.Array | npt.NDArray, "num_simplices dim dim"],
    ) -> Solution:
        """Main interface for conducting solver runs.

        The method initializes the solution vector and dispatches to the run method for the
        selected loop type.

        !!! note
            The derivator expects the metric tensor field as used in the inner product for the
            update stencil of the eikonal equation. This is the **INVERSE** of the conductivity
            tensor, which is the actual tensor field in the eikonal equation. The
            [`Tensorfield`][eikonax.tensorfield.TensorField] component provides the inverse tensor
            field.

        Args:
            tensor_field (jax.Array): Parameter field for which to solve the Eikonal equation.
                Provides an anisotropy tensor for each simplex of the mesh.

        Raises:
            ValueError: Checks that the chosen loop type is valid.

        Returns:
            Solution: Eikonax solution object.
        """
        tensor_field = jnp.array(tensor_field, dtype=jnp.float32)
        initial_guess = jnp.ones(self._vertices.shape[0]) * self._max_value
        initial_guess = initial_guess.at[self._initial_site_inds].set(self._initial_site_values)

        match self._loop_type:
            case "jitted_for":
                run_function = self._run_jitted_for_loop
            case "jitted_while":
                run_function = self._run_jitted_while_loop
            case "nonjitted_while":
                run_function = self._run_nonjitted_while_loop
            case _:
                raise ValueError(f"Invalid loop type: {self._.loop_type}")

        solution_vector, num_iterations, tolerance = run_function(initial_guess, tensor_field)
        assert solution_vector.shape[0] == self._vertices.shape[0], (
            f"Solution vector needs to have shape {self._vertices.shape[0]}, "
            f"but has shape {solution_vector.shape}"
        )
        solution = Solution(
            values=solution_vector, num_iterations=num_iterations, tolerance=tolerance
        )
        return solution

    # ----------------------------------------------------------------------------------------------
    @eqx.filter_jit
    def _run_jitted_for_loop(
        self,
        initial_guess: jtFloat[jax.Array, "num_vertices"],
        tensor_field: jtFloat[jax.Array, "num_vertices dim dim"],
    ) -> tuple[jtFloat[jax.Array, "num_vertices"], int, float]:
        """Solver run with jitted for loop for iterations.

        The method constructs a JAX-type for loop with fixed number of iterations. For every
        iteration, a new solution vector is computed from the
        [`_compute_global_update`][eikonax.solver.Solver._compute_global_update] method.

        Args:
            initial_guess (jax.Array): Initial solution vector
            tensor_field (jax.Array): Parameter field

        Returns:
            tuple[jax.Array, int, float]: Solution values, number of iterations, tolerance
        """

        # JAX body for for loop, has to carry over all args
        def loop_body_for(_: jtInt[jax.Array, ""], carry_args: tuple) -> tuple:
            new_solution_vector, tolerance, old_solution_vector, tensor_field = carry_args
            old_solution_vector = new_solution_vector
            new_solution_vector = self._compute_global_update(old_solution_vector, tensor_field)
            tolerance = jnp.max(jnp.abs(new_solution_vector - old_solution_vector))
            return new_solution_vector, tolerance, old_solution_vector, tensor_field

        initial_tolerance = 0
        initial_old_solution = jnp.zeros(initial_guess.shape)
        solution_vector, tolerance, *_ = jax.lax.fori_loop(
            0,
            self._max_num_iterations,
            loop_body_for,
            (
                initial_guess,
                initial_tolerance,
                initial_old_solution,
                tensor_field,
            ),
        )
        return solution_vector, self._max_num_iterations, tolerance

    # ----------------------------------------------------------------------------------------------
    @eqx.filter_jit
    def _run_jitted_while_loop(
        self,
        initial_guess: jtFloat[jax.Array, "num_vertices"],
        tensor_field: jtFloat[jax.Array, "num_vertices dim dim"],
    ) -> tuple[jtFloat[jax.Array, "num_vertices"], int, float]:
        """Solver run with jitted while loop for iterations.

        The iterator is tolerance-based, terminating after a user-defined tolerance for the
        difference between two consecutive iterates in supremum norm is undercut. For every
        iteration, a new solution vector is computed from the
        [`_compute_global_update`][eikonax.solver.Solver._compute_global_update] method.

        Args:
            initial_guess (jax.Array): Initial solution vector
            tensor_field (jax.Array): Parameter field

        Raises:
            ValueError: Checks that tolerance has been provided by the user

        Returns:
            tuple[jax.Array, int, float]: Solution values, number of iterations, tolerance
        """
        if self._tolerance is None:
            raise ValueError("Tolerance threshold must be provided for while loop")

        # JAX body for while loop, has to carry over all args
        def loop_body_while(carry_args: tuple) -> tuple:
            new_solution_vector, iteration_counter, tolerance, old_solution_vector, tensor_field = (
                carry_args
            )
            old_solution_vector = new_solution_vector
            new_solution_vector = self._compute_global_update(old_solution_vector, tensor_field)
            tolerance = jnp.max(jnp.abs(new_solution_vector - old_solution_vector))
            iteration_counter += 1
            return (
                new_solution_vector,
                iteration_counter,
                tolerance,
                old_solution_vector,
                tensor_field,
            )

        # JAX termination condition for while loop
        def cond_while(carry_args: tuple) -> jtBool[jax.Array, ""]:
            new_solution_vector, iteration_counter, tolerance, old_solution_vector, _ = carry_args
            tolerance = jnp.max(jnp.abs(new_solution_vector - old_solution_vector))
            return (tolerance > self._tolerance) & (iteration_counter < self._max_num_iterations)

        initial_old_solution = jnp.zeros(initial_guess.shape)
        initial_tolerance = 0
        iteration_counter = 0
        solution_vector, num_iterations, tolerance, *_ = jax.lax.while_loop(
            cond_while,
            loop_body_while,
            (
                initial_guess,
                iteration_counter,
                initial_tolerance,
                initial_old_solution,
                tensor_field,
            ),
        )
        return solution_vector, num_iterations, tolerance

    # ----------------------------------------------------------------------------------------------
    def _run_nonjitted_while_loop(
        self,
        initial_guess: jtFloat[jax.Array, "num_vertices"],
        tensor_field: jtFloat[jax.Array, "num_vertices dim dim"],
    ) -> tuple[jtFloat[jax.Array, "num_vertices"], int, jtFloat[jax.Array, "..."]]:
        """Solver run with standard Python while loop for iterations.

        While being less performant, the Python while loop allows for logging of infos between
        iterations. The iterator is tolerance-based, terminating after a user-defined tolerance for
        the difference between two consecutive iterates in supremum norm is undercut. For every
        iteration, a new solution vector is computed from the
        [`_compute_global_update`][eikonax.solver.Solver._compute_global_update] method.

        Args:
            initial_guess (jax.Array): Initial solution vector
            tensor_field (jax.Array): Parameter field

        Raises:
            ValueError: Checks that tolerance has been provided by the user
            ValueError: Checks that log interval has been provided by the user
            ValueError: Checks that logger object has been provided by the user

        Returns:
            tuple[jax.Array, int, int]: Solution values, number of iterations, tolerance
                vector over all iterations
        """
        if self._tolerance is None:
            raise ValueError("Tolerance threshold must be provided for while loop")
        if self._log_interval is None:
            raise ValueError("Log interval must be provided for non-jitted while loop")
        if self._logger is None:
            raise ValueError("Logger must be provided for non-jitted while loop")

        iteration_counter = 0
        old_solution_vector = initial_guess
        tolerance = jnp.inf
        tolerance_vector = []
        start_time = time.time()

        log_values = {
            "time": logging.LogValue(f"{'Time[s]:':<15}", "<15.3e"),
            "iters": logging.LogValue(f"{'#Iterations:':<15}", "<15.3e"),
            "tol": logging.LogValue(f"{'Tolerance:':<15}", "<15.3e"),
        }
        self._logger.header(log_values.values())

        while (tolerance > self._tolerance) and (iteration_counter < self._max_num_iterations):
            new_solution_vector = self._compute_global_update(old_solution_vector, tensor_field)
            tolerance = jnp.max(jnp.abs(new_solution_vector - old_solution_vector))
            tolerance_vector.append(tolerance)
            old_solution_vector = new_solution_vector
            iteration_counter += 1

            if (iteration_counter % self._log_interval == 0) or (
                iteration_counter == self._max_num_iterations
            ):
                current_time = time.time() - start_time
                log_values["time"].value = current_time
                log_values["iters"].value = iteration_counter
                log_values["tol"].value = tolerance
                self._logger.log(log_values.values())

        tolerance_vector = jnp.array(tolerance_vector)
        return new_solution_vector, iteration_counter, tolerance_vector

    # ----------------------------------------------------------------------------------------------
    @eqx.filter_jit
    def _compute_global_update(
        self,
        solution_vector: jtFloat[jax.Array, "num_vertices"],
        tensor_field: jtFloat[jax.Array, "num_vertices dim dim"],
    ) -> jtFloat[jax.Array, "num_vertices"]:
        """Given a current state and tensor field, compute a new solution vector.

        This method is basically a vectorized call to the
        [`_compute_vertex_update`][eikonax.solver.Solver._compute_vertex_update] method, evaluated
        over all vertices of the mesh.

        Args:
            solution_vector (jax.Array): Current state
            tensor_field (jax.Array): Parameter field

        Returns:
            jax.Array: New iterate
        """
        assert solution_vector.shape[0] == self._vertices.shape[0], (
            f"Solution vector needs to have shape {self._vertices.shape[0]}, "
            f"but has shape {solution_vector.shape}"
        )
        global_update_function = jax.vmap(self._compute_vertex_update, in_axes=(None, None, 0))
        global_update = global_update_function(
            solution_vector,
            tensor_field,
            self._adjacency_data,
        )
        assert global_update.shape == solution_vector.shape, (
            f"New solution has shape {global_update.shape}, "
            f"but should have shape {solution_vector.shape}"
        )
        return global_update

    # ----------------------------------------------------------------------------------------------
    def _compute_vertex_update(
        self,
        old_solution_vector: jtFloat[jax.Array, "num_vertices"],
        tensor_field: jtFloat[jax.Array, "num_vertices dim dim"],
        adjacency_data: jtInt[jax.Array, "max_num_adjacent_simplices 4"],
    ) -> jtFloat[jax.Array, ""]:
        """Compute the update value for a single vertex.

        This method links to the main logic of the solver routine, based on functions in the
        [`corefunctions`][eikonax.corefunctions] module.

        Args:
            old_solution_vector (jax.Array): Current state
            tensor_field (jax.Array): Parameter field
            adjacency_data (jax.Array): Info on all adjacent triangles and respective vertices
                for the current vertex

        Returns:
            jax.Array: Optimal update value for the current vertex
        """
        vertex_update_candidates = corefunctions.compute_vertex_update_candidates(
            old_solution_vector,
            tensor_field,
            adjacency_data,
            self._vertices,
            self._use_soft_update,
            self._softminmax_order,
            self._softminmax_cutoff,
        )
        self_update = jnp.expand_dims(old_solution_vector[adjacency_data[0, 0]], axis=-1)
        vertex_update_candidates = jnp.concatenate(
            (self_update, vertex_update_candidates.flatten())
        )
        vertex_update = jnp.min(vertex_update_candidates)
        assert vertex_update.shape == (), (
            f"Vertex update has to be scalar, but has shape {vertex_update.shape}"
        )
        return vertex_update



"""Test mesh creation and preparation for Eikonax solver runs.

!!! info
    The creation of test meshes can be done with any other tool. The format of the required
    adjacency data for Eikonax is strict, however.

Classes:
    MeshData: Data characterizing a computational mesh from a vertex-centered perspective.
    InitialSites: Initial site info.

Functions:
    create_test_mesh: Create a simple test mesh with Scipy's Delauny functionality.
    get_adjacent_vertex_data: Preprocess mesh data for a vertex-centered evaluation.
"""

from collections.abc import Iterable
from dataclasses import dataclass, field
from numbers import Real
from typing import Annotated

import jax
import jax.numpy as jnp
import numpy as np
import numpy.typing as npt
from beartype.vale import Is
from jaxtyping import Float as jtFloat
from jaxtyping import Int as jtInt
from scipy.spatial import Delaunay


# ==================================================================================================
@dataclass
class MeshData:
    """Data characterizing a computational mesh from a vertex-centered perspective.

    Attributes:
        vertices (jax.Array | npt.NDArray): The coordinates of the vertices in the mesh.
            The dimension of this array is `(num_vertices, dim)`, where num_vertices is the number
            of vertices in the mesh and dim is the dimension of the space in which the mesh is
            embedded.
        simplices (jax.Array | npt.NDArray): The vertex indices for each simplex in the mesh.
            The dimension of this array is `(num_simplices, 3)`, where num_simplices is the number
            of simplices in the mesh.
        adjacency_data (jax.Array | npt.NDArray): Adjacency data for each vertex. This is the list
            of adjacent triangles, together with the two vertices that span the respective triangle
            with the current vertex. The dimension of this array is
            `(num_vertices, max_num_adjacent_simplices, 4)`, where max_num_adjacent_simplices is the
            maximum number of simplices that are adjacent to a vertex in the mesh. All entries have
            this maximum size, as JAX only operates on homogeneous data structures. If a vertex has
            fewer than max_num_adjacent_simplices adjacent simplices, the remaining entries are
            filled with -1.
    """

    vertices: jtFloat[jax.Array | npt.NDArray, "num_vertices dim"]
    simplices: jtInt[jax.Array | npt.NDArray, "num_simplices 3"]
    adjacency_data: jtInt[jax.Array | npt.NDArray, "num_vertices max_num_adjacent_simplices 4"] = (
        field(init=False)
    )
    num_vertices: int = field(init=False)
    num_simplices: int = field(init=False)

    def __post_init__(self) -> None:
        """Initialize data structures and convert to JAX arrays."""
        self.num_vertices = self.vertices.shape[0]
        self.num_simplices = self.simplices.shape[0]
        self.adjacency_data = get_adjacent_vertex_data(self.simplices, self.num_vertices)
        self.vertices = jnp.array(self.vertices, dtype=jnp.float32)
        self.simplices = jnp.array(self.simplices, dtype=jnp.int32)


@dataclass
class InitialSites:
    """Initial site info.

    For a unique solution of the state-constrained Eikonal equation, the solution values need to be
    given a number of initial points (at least one). Multiple initial sites need to be compatible,
    in the sense that the arrival time from another source is not smaller than the initial value
    itself.

    Attributes:
        inds (jax.Array | npt.NDArray): The indices of the nodes where the initial sites are placed.
        values (jax.Array | npt.NDArray): The values of the initial sites.
    """

    inds: jtInt[jax.Array | npt.NDArray, "num_initial_sites"] | Iterable
    values: jtFloat[jax.Array | npt.NDArray, "num_initial_sites"] | Iterable

    def __post_init__(self) -> None:
        """Convert to jax arrays."""
        self.inds = jnp.array(self.inds, dtype=jnp.int32)
        self.values = jnp.array(self.values, dtype=jnp.float32)


# ==================================================================================================
def create_test_mesh(
    mesh_bounds_x: Annotated[Iterable[Real], Is[lambda x: len(x) == 2]],
    mesh_bounds_y: Annotated[Iterable[Real], Is[lambda x: len(x) == 2]],
    num_points_x: Annotated[int, Is[lambda x: x >= 2]],
    num_points_y: Annotated[int, Is[lambda x: x >= 2]],
) -> tuple[jtFloat[npt.NDArray, "num_vertices dim"], jtInt[npt.NDArray, "num_simplices 3"]]:
    """Create a simple test mesh with Scipy's Delaunay functionality.

    This methods creates a simple square mesh with Scipy's
    [Delaunay](https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.Delaunay.html)
    triangulation.

    Args:
        mesh_bounds_x (Iterable[float, float]): Mesh bounds for x-direction
        mesh_bounds_y (Iterable[float, float]): Mesh bounds for y-direction
        num_points_x (int): Number of vertices for x-direction
        num_points_y (int): Number of vertices for y-direction

    Raises:
        ValueError: Checks that mesh bounds have correct dimension
        ValueError: Checks that mesh bounds are provided correctly
        ValueError: Checks that at least two mesh points are chosen

    Returns:
        tuple[npt.NDArray, npt.NDArray]: Array of vertex coordinates and array of simplex indices
    """
    for mesh_bounds in (mesh_bounds_x, mesh_bounds_y):
        if mesh_bounds[0] >= mesh_bounds[1]:
            raise ValueError(
                f"Lower domain bound ({mesh_bounds[0]}) must be less than upper bound"
                f"({mesh_bounds[1]})"
            )
    mesh_points_x = np.linspace(*mesh_bounds, num_points_x)
    mesh_points_y = np.linspace(*mesh_bounds, num_points_y)
    mesh_points = np.column_stack(
        (np.repeat(mesh_points_x, num_points_x), np.tile(mesh_points_y, num_points_y))
    )
    triangulation = Delaunay(mesh_points)
    return triangulation.points, triangulation.simplices


# --------------------------------------------------------------------------------------------------
def get_adjacent_vertex_data(
    simplices: jtInt[jax.Array | npt.NDArray, "num_simplices 3"],
    num_vertices: Annotated[int, Is[lambda x: x > 0]],
) -> jtInt[jax.Array, "num_vertices max_num_adjacent_simplices 4"]:
    """Preprocess mesh data for a vertex-centered evaluation.

    Standard mesh tools provide vertex coordinates and the vertex indices for each simplex.
    For the vertex-centered solution of the Eikonal equation, however, we need the adjacent
    simplices/vertices for each vertex. This method performs the necessary transformation.

    Args:
        simplices (npt.NDArray): Vertex indices for all simplices
        num_vertices (int): Number of vertices in  the mesh

    Returns:
        npt.NDArray: Array containing for each vertex the vertex and simplex indices of all
            adjacent simplices. Dimension is `(num_vertices, max_num_adjacent_simplices, 4)`,
            where the 4 entries contain the index of an adjacent simplex and the associated
            vertices. To ensure homogeneous arrays, all vertices have the same (maximum) number
            of adjacent simplices. Non-existing simplices are buffered with the value -1.
    """
    max_num_adjacent_simplices = np.max(np.bincount(simplices.flatten()))
    adjacent_vertex_inds = -1 * np.ones(
        (num_vertices, max_num_adjacent_simplices, 4), dtype=np.int32
    )
    counter_array = np.zeros(num_vertices, dtype=int)
    node_permutations = ((0, 1, 2), (1, 0, 2), (2, 0, 1))

    for simplex_inds, simplex in enumerate(simplices):
        for permutation in node_permutations:
            center_vertex, adj_vertex_1, adj_vertex_2 = simplex[permutation,]
            adjacent_vertex_inds[center_vertex, counter_array[center_vertex]] = np.array(
                [center_vertex, adj_vertex_1, adj_vertex_2, simplex_inds]
            )
            counter_array[center_vertex] += 1
    adjacent_vertex_inds = jnp.array(adjacent_vertex_inds, dtype=jnp.int32)

    return adjacent_vertex_inds



import jax.numpy as jnp
import numpy as np
import pytest

from eikonax import corefunctions, preprocessing

pytestmark = pytest.mark.unit


# =================================== Unit Tests for Preprocssing ==================================
def test_create_test_mesh(mesh_small):
    expected_vertices, expected_simplices, meta_data = mesh_small
    created_vertices, created_simplices = preprocessing.create_test_mesh(**meta_data)
    assert np.allclose(expected_vertices, created_vertices)
    assert np.allclose(expected_simplices, created_simplices)


# --------------------------------------------------------------------------------------------------
@staticmethod
def test_get_adjacent_vertex_data(mesh_and_adjacency_data_small):
    benchmark_mesh, expected_adjacency_data = mesh_and_adjacency_data_small
    benchmark_vertices, benchmark_simplices, _ = benchmark_mesh
    adjacency_data = preprocessing.get_adjacent_vertex_data(
        benchmark_simplices, benchmark_vertices.shape[0]
    )
    assert np.allclose(adjacency_data, expected_adjacency_data)


# ================================== Unit Tests for Core Functions =================================
@pytest.mark.parametrize("order", [1, 10, 1000], ids=["order_1", "order_10", "order_1000"])
@pytest.mark.parametrize(
    "input_value",
    [
        jnp.array(0),
        jnp.array(1),
        jnp.array(0.25),
        jnp.array(0.75),
        jnp.array(-1000),
        jnp.array(1000),
        jnp.array(-jnp.inf),
        jnp.array(jnp.inf),
    ],
)
def test_bounds_compute_softminmax(input_value, order):
    output = corefunctions.compute_softminmax(input_value, order)
    assert jnp.all(output >= 0)
    assert jnp.all(output <= 1)


# --------------------------------------------------------------------------------------------------
def test_values_compute_softminmax():
    order = 10
    input_values = jnp.linspace(-1, 2, 10, dtype=jnp.float32)
    expected_output_values = jnp.array(
        [
            4.5417501e-06,
            1.2718633e-04,
            3.5050693e-03,
            6.9309868e-02,
            3.3670986e-01,
            6.6328585e-01,
            9.3068784e-01,
            9.9649477e-01,
            9.9987280e-01,
            9.9999547e-01,
        ],
        dtype=jnp.float32,
    )
    output_values = corefunctions.compute_softminmax(input_values, order)
    assert jnp.allclose(output_values, expected_output_values)


# --------------------------------------------------------------------------------------------------
def test_compute_edges():
    test_vertices = jnp.array(([-1, -1], [1, 0], [0.5, 2]))
    expected_edges = (jnp.array([-2, -1]), jnp.array([-1.5, -3]), jnp.array([-0.5, 2]))
    output_edges = corefunctions.compute_edges(
        jnp.array(0), jnp.array(1), jnp.array(2), test_vertices
    )
    for output_edge, expected_edge in zip(output_edges, expected_edges, strict=True):
        assert jnp.allclose(output_edge, expected_edge)


# --------------------------------------------------------------------------------------------------
def test_compute_optimal_update_parameters(simplex_setup_and_optimal_lambda_candidates):
    *input_data, expected_lambda_values = simplex_setup_and_optimal_lambda_candidates
    output_lambda_values = corefunctions._compute_optimal_update_parameters(*input_data)
    for output_lambda, expected_lambda in zip(
        output_lambda_values, expected_lambda_values, strict=True
    ):
        assert jnp.allclose(output_lambda, expected_lambda)


# --------------------------------------------------------------------------------------------------
@pytest.mark.parametrize(
    ("lambda_input", "lambda_expected"),
    [
        ([0.25, 0.75], jnp.array((0, 1, 0.25, 0.75))),
        ([0.4, 0.4], jnp.array((0, 1, 0.4, -1))),
        ([0, 0.5], jnp.array((0, 1, -1, 0.5))),
        ([0.5, 0], jnp.array((0, 1, 0.5, -1))),
        ([-10, 0.1], jnp.array((0, 1, -1, 0.1))),
        ([0.1, 10], jnp.array((0, 1, 0.1, -1))),
        ([-10, 10], jnp.array((0, 1, -1, -1))),
    ],
)
def test_compute_optimal_update_parameters_hard(
    simplex_setup_and_optimal_lambda_candidates, lambda_input, lambda_expected, monkeypatch
):
    *input_data, _ = simplex_setup_and_optimal_lambda_candidates
    monkeypatch.setattr(
        "eikonax.corefunctions._compute_optimal_update_parameters", lambda *_: lambda_input
    )
    output_lambda_values = corefunctions.compute_optimal_update_parameters_hard(*input_data)
    assert jnp.allclose(output_lambda_values, lambda_expected)


# --------------------------------------------------------------------------------------------------
def test_compute_optimal_update_parameters_soft(
    simplex_setup_and_optimal_lambda_candidates, monkeypatch
):
    order, cutoff = 1, 1
    lambda_input = [jnp.array(-1.1, dtype=jnp.float32), jnp.array(2.1, dtype=jnp.float32)]
    *input_data, _ = simplex_setup_and_optimal_lambda_candidates
    monkeypatch.setattr(
        "eikonax.corefunctions._compute_optimal_update_parameters", lambda *_: lambda_input
    )
    output_lambda_values = corefunctions.compute_optimal_update_parameters_soft(
        *input_data, order, cutoff
    )
    assert jnp.allclose(output_lambda_values, jnp.array((0, 1, -1, -1)))


# --------------------------------------------------------------------------------------------------
def test_compute_fixed_update(simplex_setup_and_update):
    *input_data, expected_update_value = simplex_setup_and_update
    update_value = corefunctions.compute_fixed_update(*input_data)
    assert jnp.allclose(update_value, expected_update_value)


# --------------------------------------------------------------------------------------------------
def test_grad_update_solution(simplex_data_update_derivative):
    simplex_setup_and_update, expected_grad_update_solution, *_ = simplex_data_update_derivative
    *input_data, _ = simplex_setup_and_update
    grad_update_solution = corefunctions.grad_update_solution(*input_data)
    assert jnp.allclose(grad_update_solution, expected_grad_update_solution)


# --------------------------------------------------------------------------------------------------
def test_grad_update_parameter(simplex_data_update_derivative):
    simplex_setup_and_update, _, expected_grad_update_parameter, _ = simplex_data_update_derivative
    *input_data, _ = simplex_setup_and_update
    grad_update_parameter = corefunctions.grad_update_parameter(*input_data)
    assert jnp.allclose(grad_update_parameter, expected_grad_update_parameter)


# --------------------------------------------------------------------------------------------------
def test_grad_update_lambda(simplex_data_update_derivative):
    simplex_setup_and_update, *_, expected_grad_update_lambda = simplex_data_update_derivative
    *input_data, _ = simplex_setup_and_update
    grad_update_lambda = corefunctions.grad_update_lambda(*input_data)
    assert jnp.allclose(grad_update_lambda, expected_grad_update_lambda)


# --------------------------------------------------------------------------------------------------
@pytest.mark.slow
@pytest.mark.parametrize(
    "update_fixture", ["update_candidates_without_softminmax", "update_candidates_with_softminmax"]
)
def test_compute_global_update_candidates(update_fixture, request):
    update_setup, use_soft_update, solution_values, expected_update_candidates = (
        request.getfixturevalue(update_fixture)
    )
    vertices, adjacency_data, tensor_field, softminmax_order, softminmax_cutoff = update_setup
    vertices = jnp.array(vertices, dtype=jnp.float32)
    adjacency_data = jnp.array(adjacency_data, dtype=jnp.int32)

    for i in range(vertices.shape[0]):
        adj_data = adjacency_data[i, ...]
        exp_update_candidates = expected_update_candidates[i, ...]
        update_candidates = corefunctions.compute_vertex_update_candidates(
            solution_values,
            tensor_field,
            adj_data,
            vertices,
            use_soft_update,
            softminmax_order,
            softminmax_cutoff,
        )
        assert jnp.allclose(update_candidates, exp_update_candidates)
